from ._codetemplate import image_translation_prf
import json
import traceback
from .._data import _raise_fastai_import_error
from ._arcgis_model import ArcGISModel
try:
    from ._pix2pix_utils import pix2pixLoss, pix2pixTrainer, optim, compute_metrics, compute_fid_metric
    from ._pix2pix_utils import  pix2pix as pix2pix_model
    from .._utils.pix2pix import ImageTuple, ImageTupleList2, ImageTupleListMS2
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from torchvision import transforms
    from pathlib import Path
    from fastai.vision import *
    from fastai.vision import DatasetType, Learner, partial, open_image
    import torch

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False

class Pix2Pix(ArcGISModel):

    """
    Creates a model object which generates fake images of type B from type A.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================
                                             
    :returns: `Pix2Pix` Object
    """
    
    def __init__(self, data, pretrained_path=None, *args, **kwargs):
        super().__init__(data)
        pix2pix_gan = pix2pix_model(self._data.n_channel,self._data.n_channel)
        self.learn = Learner(data, 
                             pix2pix_gan, 
                             loss_func=pix2pixLoss(pix2pix_gan), 
                             opt_func=partial(optim.Adam,betas=(0.5,0.99)),
                             callback_fns=[pix2pixTrainer])

        self.learn.model = self.learn.model.to(self._device)
        self._slice_lr = False
        if pretrained_path is not None:
            self.load(pretrained_path)
        self._code = image_translation_prf
        def __str__(self):
            return self.__repr__()
        def __repr__(self):
            return '<%s>' % (type(self).__name__)
        
    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a Pix2Pix object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================
        
        :returns: `Pix2Pix` Object
        """
        
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)
            
        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd['ModelFile'])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd['ModelParameters']
        resize_to = emd.get('resize_to')
        chip_size = emd['ImageHeight']
        if data is None:
             if emd.get('IsMultispectral', False):
                data = ImageTupleListMS2.from_folders(emd_path.parent, emd_path.parent, emd_path.parent, batch_stats_a=None, batch_stats_b=None).split_none().label_empty().databunch(bs=2,no_check = True)
                data.n_channel = emd['n_channel']
                data = get_multispectral_data_params_from_emd(data, emd)
                data._is_multispectral = emd.get('IsMultispectral', False)
                normalization_stats_b = dict(emd.get("NormalizationStats_b"))
                for _stat in normalization_stats_b:
                    if normalization_stats_b[_stat] is not None:
                        normalization_stats_b[_stat] = torch.tensor(normalization_stats_b[_stat])
                    setattr(data, ('_'+_stat), normalization_stats_b[_stat])

             else:
                 data = ImageTupleList2.from_folders(emd_path.parent, emd_path.parent, emd_path.parent)\
                     .split_none()\
                     .label_empty()\
                     .transform(size=(chip_size, chip_size))\
                     .databunch(bs=2, no_check = True)
        data.n_channel = emd['n_channel']
        data._is_empty = True
        data.emd_path = emd_path
        data.emd = emd
        data.resize_to = chip_size
        
        return cls(data, **model_params, pretrained_path=str(model_file))
        
    @property
    def _model_metrics(self):
        psnr_ssim_fid = self.compute_metrics(show_progress=True)
        return {'psnr_metric': '{}'.format(psnr_ssim_fid["PSNR"]),
                'ssim_metric': '{}'.format(psnr_ssim_fid["SSIM"]),
                'FID': '{}'.format(psnr_ssim_fid["FID"])}

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_pix2pix"
        _emd_template["InferenceFunction"] = "ArcGISImageTranslation.py"
        _emd_template["ModelType"] = "Pix2Pix"
        _emd_template["n_channel"] = self._data.n_channel
        if self._data._is_multispectral:
            _emd_template["NormalizationStats_b"] = {
                    "band_min_values": self._data._band_min_values_b,
                    "band_max_values": self._data._band_max_values_b,
                    "band_mean_values": self._data._band_mean_values_b,
                    "band_std_values": self._data._band_std_values_b,
                    "scaled_min_values": self._data._scaled_min_values_b,
                    "scaled_max_values": self._data._scaled_max_values_b,
                    "scaled_mean_values": self._data._scaled_mean_values_b,
                    "scaled_std_values": self._data._scaled_std_values_b
        }
            for _stat in _emd_template["NormalizationStats_b"]:
                    if _emd_template["NormalizationStats_b"][_stat] is not None:
                        _emd_template["NormalizationStats_b"][_stat] = _emd_template["NormalizationStats_b"][_stat].tolist()
        return _emd_template

    def show_results(self,rows=5):
        """
        Displays the results of a trained model on a part of the validation set.

        """

        self.learn.model.arcgis_results = True
        self.learn.show_results()
        self.learn.model.arcgis_results = False

    def predict(self, img_path):
        """
        Predicts and display the image.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        img_path                Required path of an image.
        =====================   ===========================================

        """
        self.learn.model.arcgis_results = True
        img_path = Path(img_path)
        raw_img = open_image(img_path)
        n_band = self._data.n_channel
        if n_band > raw_img.shape[0]:
            cont = []
            last_tile = np.expand_dims(raw_img.data[raw_img.shape[0]-1,:,:], 0)
            res = abs(n_band - raw_img.shape[0])
            for i in range(res):
                raw_img = Image(torch.tensor(np.concatenate((raw_img.data, last_tile), axis=0)))
        raw_img_tuple = ImageTuple(raw_img, raw_img)
        pred_tuple = self.learn.predict(raw_img_tuple)
        pred_img = pred_tuple[1][0]/2+0.5
        
        pred_img = transforms.ToPILImage()(pred_img).convert("RGB")
        self.learn.model.arcgis_results = False
        return pred_img

    def compute_metrics(self, accuracy=True, show_progress=True):
        """
        Computes Peak Signal-to-Noise Ratio (PSNR) and 
        Structural Similarity Index Measure (SSIM) on validation set.

        """
        psnr, ssim = compute_metrics(self, self._data.valid_dl, show_progress)
        if self._data.n_channel > 3:
            fid = None
            return {"PSNR":psnr, "SSIM":ssim}
        else:
            fid = compute_fid_metric(self, self._data)
            return {"PSNR":psnr, "SSIM":ssim, "FID":fid}



# Apache License
# Version 2.0, January 2004
# http://www.apache.org/licenses/

# TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

# 1. Definitions.

# "License" shall mean the terms and conditions for use, reproduction,
# and distribution as defined by Sections 1 through 9 of this document.

# "Licensor" shall mean the copyright owner or entity authorized by
# the copyright owner that is granting the License.

# "Legal Entity" shall mean the union of the acting entity and all
# other entities that control, are controlled by, or are under common
# control with that entity. For the purposes of this definition,
# "control" means (i) the power, direct or indirect, to cause the
# direction or management of such entity, whether by contract or
# otherwise, or (ii) ownership of fifty percent (50%) or more of the
# outstanding shares, or (iii) beneficial ownership of such entity.

# "You" (or "Your") shall mean an individual or Legal Entity
# exercising permissions granted by this License.

# "Source" form shall mean the preferred form for making modifications,
# including but not limited to software source code, documentation
# source, and configuration files.

# "Object" form shall mean any form resulting from mechanical
# transformation or translation of a Source form, including but
# not limited to compiled object code, generated documentation,
# and conversions to other media types.

# "Work" shall mean the work of authorship, whether in Source or
# Object form, made available under the License, as indicated by a
# copyright notice that is included in or attached to the work
# (an example is provided in the Appendix below).

# "Derivative Works" shall mean any work, whether in Source or Object
# form, that is based on (or derived from) the Work and for which the
# editorial revisions, annotations, elaborations, or other modifications
# represent, as a whole, an original work of authorship. For the purposes
# of this License, Derivative Works shall not include works that remain
# separable from, or merely link (or bind by name) to the interfaces of,
# the Work and Derivative Works thereof.

# "Contribution" shall mean any work of authorship, including
# the original version of the Work and any modifications or additions
# to that Work or Derivative Works thereof, that is intentionally
# submitted to Licensor for inclusion in the Work by the copyright owner
# or by an individual or Legal Entity authorized to submit on behalf of
# the copyright owner. For the purposes of this definition, "submitted"
# means any form of electronic, verbal, or written communication sent
# to the Licensor or its representatives, including but not limited to
# communication on electronic mailing lists, source code control systems,
# and issue tracking systems that are managed by, or on behalf of, the
# Licensor for the purpose of discussing and improving the Work, but
# excluding communication that is conspicuously marked or otherwise
# designated in writing by the copyright owner as "Not a Contribution."

# "Contributor" shall mean Licensor and any individual or Legal Entity
# on behalf of whom a Contribution has been received by Licensor and
# subsequently incorporated within the Work.

# 2. Grant of Copyright License. Subject to the terms and conditions of
# this License, each Contributor hereby grants to You a perpetual,
# worldwide, non-exclusive, no-charge, royalty-free, irrevocable
# copyright license to reproduce, prepare Derivative Works of,
# publicly display, publicly perform, sublicense, and distribute the
# Work and such Derivative Works in Source or Object form.

# 3. Grant of Patent License. Subject to the terms and conditions of
# this License, each Contributor hereby grants to You a perpetual,
# worldwide, non-exclusive, no-charge, royalty-free, irrevocable
# (except as stated in this section) patent license to make, have made,
# use, offer to sell, sell, import, and otherwise transfer the Work,
# where such license applies only to those patent claims licensable
# by such Contributor that are necessarily infringed by their
# Contribution(s) alone or by combination of their Contribution(s)
# with the Work to which such Contribution(s) was submitted. If You
# institute patent litigation against any entity (including a
# cross-claim or counterclaim in a lawsuit) alleging that the Work
# or a Contribution incorporated within the Work constitutes direct
# or contributory patent infringement, then any patent licenses
# granted to You under this License for that Work shall terminate
# as of the date such litigation is filed.

# 4. Redistribution. You may reproduce and distribute copies of the
# Work or Derivative Works thereof in any medium, with or without
# modifications, and in Source or Object form, provided that You
# meet the following conditions:

# (a) You must give any other recipients of the Work or
# Derivative Works a copy of this License; and

# (b) You must cause any modified files to carry prominent notices
# stating that You changed the files; and

# (c) You must retain, in the Source form of any Derivative Works
# that You distribute, all copyright, patent, trademark, and
# attribution notices from the Source form of the Work,
# excluding those notices that do not pertain to any part of
# the Derivative Works; and

# (d) If the Work includes a "NOTICE" text file as part of its
# distribution, then any Derivative Works that You distribute must
# include a readable copy of the attribution notices contained
# within such NOTICE file, excluding those notices that do not
# pertain to any part of the Derivative Works, in at least one
# of the following places: within a NOTICE text file distributed
# as part of the Derivative Works; within the Source form or
# documentation, if provided along with the Derivative Works; or,
# within a display generated by the Derivative Works, if and
# wherever such third-party notices normally appear. The contents
# of the NOTICE file are for informational purposes only and
# do not modify the License. You may add Your own attribution
# notices within Derivative Works that You distribute, alongside
# or as an addendum to the NOTICE text from the Work, provided
# that such additional attribution notices cannot be construed
# as modifying the License.

# You may add Your own copyright statement to Your modifications and
# may provide additional or different license terms and conditions
# for use, reproduction, or distribution of Your modifications, or
# for any such Derivative Works as a whole, provided Your use,
# reproduction, and distribution of the Work otherwise complies with
# the conditions stated in this License.

# 5. Submission of Contributions. Unless You explicitly state otherwise,
# any Contribution intentionally submitted for inclusion in the Work
# by You to the Licensor shall be under the terms and conditions of
# this License, without any additional terms or conditions.
# Notwithstanding the above, nothing herein shall supersede or modify
# the terms of any separate license agreement you may have executed
# with Licensor regarding such Contributions.

# 6. Trademarks. This License does not grant permission to use the trade
# names, trademarks, service marks, or product names of the Licensor,
# except as required for reasonable and customary use in describing the
# origin of the Work and reproducing the content of the NOTICE file.

# 7. Disclaimer of Warranty. Unless required by applicable law or
# agreed to in writing, Licensor provides the Work (and each
# Contributor provides its Contributions) on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied, including, without limitation, any warranties or conditions
# of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
# PARTICULAR PURPOSE. You are solely responsible for determining the
# appropriateness of using or redistributing the Work and assume any
# risks associated with Your exercise of permissions under this License.

# 8. Limitation of Liability. In no event and under no legal theory,
# whether in tort (including negligence), contract, or otherwise,
# unless required by applicable law (such as deliberate and grossly
# negligent acts) or agreed to in writing, shall any Contributor be
# liable to You for damages, including any direct, indirect, special,
# incidental, or consequential damages of any character arising as a
# result of this License or out of the use or inability to use the
# Work (including but not limited to damages for loss of goodwill,
# work stoppage, computer failure or malfunction, or any and all
# other commercial damages or losses), even if such Contributor
# has been advised of the possibility of such damages.

# 9. Accepting Warranty or Additional Liability. While redistributing
# the Work or Derivative Works thereof, You may choose to offer,
# and charge a fee for, acceptance of support, warranty, indemnity,
# or other liability obligations and/or rights consistent with this
# License. However, in accepting such obligations, You may act only
# on Your own behalf and on Your sole responsibility, not on behalf
# of any other Contributor, and only if You agree to indemnify,
# defend, and hold each Contributor harmless for any liability
# incurred by, or claims asserted against, such Contributor by reason
# of your accepting any such warranty or additional liability.

# END OF TERMS AND CONDITIONS

# APPENDIX: How to apply the Apache License to your work.

# To apply the Apache License to your work, attach the following
# boilerplate notice, with the fields enclosed by brackets "[]"
# replaced with your own identifying information. (Don't include
# the brackets!)  The text should be enclosed in the appropriate
# comment syntax for the file format. We also recommend that a
# file or class name and description of purpose be included on the
# same "printed page" as the copyright notice for easier
# identification within third-party archives.

# Copyright 2019 - present, Facebook, Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#Based on https://github.com/facebookresearch/detectron2/blob/master/projects/PointRend

import numpy as np
from typing import Dict
import torch
from torch import nn
from torch.nn import functional as F

def calculate_uncertainty(sem_seg_logits):
    """
    For each location of the prediction `sem_seg_logits` we estimate uncerainty as the
        difference between top first and top second predicted logits.

    Args:
        mask_logits (Tensor): A tensor of shape (N, C, ...), where N is the minibatch size and
            C is the number of foreground classes. The values are logits.

    Returns:
        scores (Tensor): A tensor of shape (N, 1, ...) that contains uncertainty scores with
            the most uncertain locations having the highest uncertainty score.
    """
    top2_scores = torch.topk(sem_seg_logits, k=2, dim=1)[0]
    return (top2_scores[:, 1] - top2_scores[:, 0]).unsqueeze(1)

def point_sample(input, point_coords, **kwargs):
    """
    A wrapper around :function:`torch.nn.functional.grid_sample` to support 3D point_coords tensors.
    Unlike :function:`torch.nn.functional.grid_sample` it assumes `point_coords` to lie inside
    [0, 1] x [0, 1] square.

    Args:
        input (Tensor): A tensor of shape (N, C, H, W) that contains features map on a H x W grid.
        point_coords (Tensor): A tensor of shape (N, P, 2) or (N, Hgrid, Wgrid, 2) that contains
        [0, 1] x [0, 1] normalized point coordinates.

    Returns:
        output (Tensor): A tensor of shape (N, C, P) or (N, C, Hgrid, Wgrid) that contains
            features for points in `point_coords`. The features are obtained via bilinear
            interplation from `input` the same way as :function:`torch.nn.functional.grid_sample`.
    """
    add_dim = False
    if point_coords.dim() == 3:
        add_dim = True
        point_coords = point_coords.unsqueeze(2)
    output = F.grid_sample(input, 2.0 * point_coords - 1.0, **kwargs)
    if add_dim:
        output = output.squeeze(3)
    return output

def get_uncertain_point_coords_on_grid(uncertainty_map, num_points):
    """
    Find `num_points` most uncertain points from `uncertainty_map` grid.

    Args:
        uncertainty_map (Tensor): A tensor of shape (N, 1, H, W) that contains uncertainty
            values for a set of points on a regular H x W grid.
        num_points (int): The number of points P to select.

    Returns:
        point_indices (Tensor): A tensor of shape (N, P) that contains indices from
            [0, H x W) of the most uncertain points.
        point_coords (Tensor): A tensor of shape (N, P, 2) that contains [0, 1] x [0, 1] normalized
            coordinates of the most uncertain points from the H x W grid.
    """
    R, _, H, W = uncertainty_map.shape
    h_step = 1.0 / float(H)
    w_step = 1.0 / float(W)

    num_points = min(H * W, num_points)
    point_indices = torch.topk(uncertainty_map.view(R, H * W), k=num_points, dim=1)[1]
    point_coords = torch.zeros(R, num_points, 2, dtype=torch.float, device=uncertainty_map.device)
    point_coords[:, :, 0] = w_step / 2.0 + (point_indices % W).to(torch.float) * w_step
    point_coords[:, :, 1] = h_step / 2.0 + (point_indices // W).to(torch.float) * h_step
    return point_indices, point_coords

def get_uncertain_point_coords_with_randomness(
    coarse_logits, uncertainty_func, num_points, oversample_ratio, importance_sample_ratio
):
    """
    Sample points in [0, 1] x [0, 1] coordinate space based on their uncertainty. The unceratinties
        are calculated for each point using 'uncertainty_func' function that takes point's logit
        prediction as input.
    See PointRend paper for details.

    Args:
        coarse_logits (Tensor): A tensor of shape (N, C, Hmask, Wmask) or (N, 1, Hmask, Wmask) for
            class-specific or class-agnostic prediction.
        uncertainty_func: A function that takes a Tensor of shape (N, C, P) or (N, 1, P) that
            contains logit predictions for P points and returns their uncertainties as a Tensor of
            shape (N, 1, P).
        num_points (int): The number of points P to sample.
        oversample_ratio (int): Oversampling parameter.
        importance_sample_ratio (float): Ratio of points that are sampled via importnace sampling.

    Returns:
        point_coords (Tensor): A tensor of shape (N, P, 2) that contains the coordinates of P
            sampled points.
    """
    assert oversample_ratio >= 1
    assert importance_sample_ratio <= 1 and importance_sample_ratio >= 0
    num_boxes = coarse_logits.shape[0]
    num_sampled = int(num_points * oversample_ratio)
    point_coords = torch.rand(num_boxes, num_sampled, 2, device=coarse_logits.device)
    point_logits = point_sample(coarse_logits, point_coords, align_corners=False)
    # It is crucial to calculate uncertainty based on the sampled prediction value for the points.
    # Calculating uncertainties of the coarse predictions first and sampling them for points leads
    # to incorrect results.
    # To illustrate this: assume uncertainty_func(logits)=-abs(logits), a sampled point between
    # two coarse predictions with -1 and 1 logits has 0 logits, and therefore 0 uncertainty value.
    # However, if we calculate uncertainties for the coarse predictions first,
    # both will have -1 uncertainty, and the sampled point will get -1 uncertainty.
    point_uncertainties = uncertainty_func(point_logits)
    num_uncertain_points = int(importance_sample_ratio * num_points)
    num_random_points = num_points - num_uncertain_points
    idx = torch.topk(point_uncertainties[:, 0, :], k=num_uncertain_points, dim=1)[1]
    shift = num_sampled * torch.arange(num_boxes, dtype=torch.long, device=coarse_logits.device)
    idx += shift[:, None]
    point_coords = point_coords.view(-1, 2)[idx.view(-1), :].view(
        num_boxes, num_uncertain_points, 2
    )
    if num_random_points > 0:
        point_coords = torch.cat(
            [
                point_coords,
                torch.rand(num_boxes, num_random_points, 2, device=coarse_logits.device),
            ],
            dim=1,
        )
    return point_coords


def generate_regular_grid_point_coords(R, side_size, device):
    """
    Generate regular square grid of points in [0, 1] x [0, 1] coordinate space.

    Args:
        R (int): The number of grids to sample, one for each region.
        side_size (int): The side size of the regular grid.
        device (torch.device): Desired device of returned tensor.

    Returns:
        (Tensor): A tensor of shape (R, side_size^2, 2) that contains coordinates
            for the regular grids.
    """
    aff = torch.tensor([[[0.5, 0, 0.5], [0, 0.5, 0.5]]], device=device)
    r = F.affine_grid(aff, torch.Size((1, 1, side_size, side_size)), align_corners=False)
    return r.view(1, -1, 2).expand(R, -1, -1)

def point_sample_fine_grained_features(features_list, feature_scales, boxes, point_coords):
    """
    Get features from feature maps in `features_list` that correspond to specific point coordinates
        inside each bounding box from `boxes`.

    Args:
        features_list (list[Tensor]): A list of feature map tensors to get features from.
        feature_scales (list[float]): A list of scales for tensors in `features_list`.
        boxes (list[Boxes]): A list of I Boxes  objects that contain R_1 + ... + R_I = R boxes all
            together.
        point_coords (Tensor): A tensor of shape (R, P, 2) that contains
            [0, 1] x [0, 1] box-normalized coordinates of the P sampled points.

    Returns:
        point_features (Tensor): A tensor of shape (R, C, P) that contains features sampled
            from all features maps in feature_list for P sampled points for all R boxes in `boxes`.
        point_coords_wrt_image (Tensor): A tensor of shape (R, P, 2) that contains image-level
            coordinates of P points.
    """
    cat_boxes = torch.cat(boxes, dim=0)
    num_boxes = [len(b) for b in boxes]
    point_coords_wrt_image = get_point_coords_wrt_image(cat_boxes, point_coords)
    split_point_coords_wrt_image = torch.split(point_coords_wrt_image, num_boxes)
    point_features = []
    for idx_img, point_coords_wrt_image_per_image in enumerate(split_point_coords_wrt_image):
        point_features_per_image = []
        for idx_feature, feature_map in enumerate(features_list):
            h, w = feature_map.shape[-2:]
            scale = torch.tensor([w, h]) / feature_scales[idx_feature]
            scale = scale.to(feature_map.device)
            point_coords_scaled = point_coords_wrt_image_per_image / scale
            point_features_per_image.append(
                point_sample(
                    feature_map[idx_img].unsqueeze(0),
                    point_coords_scaled.unsqueeze(0),
                    align_corners=False,
                )
                .squeeze(0)
                .transpose(1, 0)
            )
        point_features.append(torch.cat(point_features_per_image, dim=1))

    return torch.cat(point_features, dim=0), point_coords_wrt_image

def get_point_coords_wrt_image(boxes_coords, point_coords):
    """
    Convert box-normalized [0, 1] x [0, 1] point cooordinates to image-level coordinates.

    Args:
        boxes_coords (Tensor): A tensor of shape (R, 4) that contains bounding boxes.
            coordinates.
        point_coords (Tensor): A tensor of shape (R, P, 2) that contains
            [0, 1] x [0, 1] box-normalized coordinates of the P sampled points.

    Returns:
        point_coords_wrt_image (Tensor): A tensor of shape (R, P, 2) that contains
            image-normalized coordinates of P sampled points.
    """
    with torch.no_grad():
        point_coords_wrt_image = point_coords.clone()
        point_coords_wrt_image[:, :, 0] = point_coords_wrt_image[:, :, 0] * (
            boxes_coords[:, None, 2] - boxes_coords[:, None, 0]
        )
        point_coords_wrt_image[:, :, 1] = point_coords_wrt_image[:, :, 1] * (
            boxes_coords[:, None, 3] - boxes_coords[:, None, 1]
        )
        point_coords_wrt_image[:, :, 0] += boxes_coords[:, None, 0]
        point_coords_wrt_image[:, :, 1] += boxes_coords[:, None, 1]
    return point_coords_wrt_image

def roi_mask_point_loss(mask_logits, instances, points_coord):
    """
    Compute the point-based loss for instance segmentation mask predictions.

    Args:
        mask_logits (Tensor): A tensor of shape (R, C, P) or (R, 1, P) for class-specific or
            class-agnostic, where R is the total number of predicted masks in all images, C is the
            number of foreground classes, and P is the number of points sampled for each mask.
            The values are logits.
        instances (list[Instances]): A list of N Instances, where N is the number of images
            in the batch. These instances are in 1:1 correspondence with the `mask_logits`. So, i_th
            elememt of the list contains R_i objects and R_1 + ... + R_N is equal to R.
            The ground-truth labels (class, box, mask, ...) associated with each instance are stored
            in fields.
        points_coords (Tensor): A tensor of shape (R, P, 2), where R is the total number of
            predicted masks and P is the number of points for each mask. The coordinates are in
            the image pixel coordinate space, i.e. [0, H] x [0, W].
    Returns:
        point_loss (Tensor): A scalar tensor containing the loss.
    """
    with torch.no_grad():
        cls_agnostic_mask = mask_logits.size(1) == 1
        total_num_masks = mask_logits.size(0)

        gt_classes = []
        gt_mask_logits = []
        idx = 0
        for instances_per_image in instances:
            if instances_per_image['gt_masks'].shape[0] == 0:
                continue

            if not cls_agnostic_mask:
                gt_classes_per_image = instances_per_image['gt_classes'].to(dtype=torch.int64)
                gt_classes.append(gt_classes_per_image)

            gt_bit_masks = instances_per_image['gt_masks']
            h, w = instances_per_image['gt_masks'].shape[-2:]#instances_per_image.gt_masks.image_size
            scale = torch.tensor([w, h], dtype=torch.float, device=gt_bit_masks.device)
            points_coord_grid_sample_format = (
                points_coord[idx : idx + instances_per_image['gt_masks'].shape[0]] / scale
            )
            idx += instances_per_image['gt_masks'].shape[0]
            gt_mask_logits.append(
                point_sample(
                    gt_bit_masks.to(torch.float32).unsqueeze(1),
                    points_coord_grid_sample_format,
                    align_corners=False,
                ).squeeze(1)
            )

    if len(gt_mask_logits) == 0:
        return mask_logits.sum() * 0

    gt_mask_logits = torch.cat(gt_mask_logits)
    assert gt_mask_logits.numel() > 0, gt_mask_logits.shape

    if cls_agnostic_mask:
        mask_logits = mask_logits[:, 0]
    else:
        indices = torch.arange(total_num_masks)
        gt_classes = torch.cat(gt_classes, dim=0)
        mask_logits = mask_logits[indices, gt_classes]

    point_loss = F.binary_cross_entropy_with_logits(
        mask_logits, gt_mask_logits.to(dtype=torch.float32), reduction="mean"
    )
    return point_loss

def PointRend_target_transform(targets, point_coords):

    point_targets = (
        point_sample(
            targets.unsqueeze(1).to(torch.float),
            point_coords,
            mode="nearest",
            align_corners=False,
        )
        .squeeze(1)
        .to(torch.long)
    )

    return point_targets

def c2_msra_fill(module: nn.Module):
    """
    Initialize `module.weight` using the "MSRAFill" implemented in Caffe2.
    Also initializes `module.bias` to 0.
    Args:
        module (torch.nn.Module): module to initialize.
    """
    # pyre-ignore
    nn.init.kaiming_normal_(module.weight, mode="fan_out", nonlinearity="relu")
    if module.bias is not None:  # pyre-ignore
        nn.init.constant_(module.bias, 0)

def c2_xavier_fill(module: nn.Module):
    """
    Initialize `module.weight` using the "XavierFill" implemented in Caffe2.
    Also initializes `module.bias` to 0.
    Args:
        module (torch.nn.Module): module to initialize.
    """
    # Caffe2 implementation of XavierFill in fact
    # corresponds to kaiming_uniform_ in PyTorch
    nn.init.kaiming_uniform_(module.weight, a=1)  # pyre-ignore
    if module.bias is not None:  # pyre-ignore
        nn.init.constant_(module.bias, 0)

class PointRendSemSegHead(nn.Module):
    """
    A semantic segmentation head that combines a head set in `POINT_HEAD.COARSE_SEM_SEG_HEAD_NAME`
    and a point head set in `MODEL.POINT_HEAD.NAME`.
    """

    def __init__(self, num_classes, backbone_features_channel, **kwargs):
        super().__init__()
        self.train_num_points        = max(1024, int(kwargs.get('train_num_points', 1024)))
        self.oversample_ratio        = 3
        self.importance_sample_ratio = 0.75
        self.subdivision_steps       = int(kwargs.get('subdivision_steps', 3))
        self.subdivision_num_points  = int(kwargs.get('subdivision_num_points', 1024))
        self.point_head = StandardPointHead(num_classes, backbone_features_channel)

    def forward(self, coarse_sem_seg_logits, backbone_features):

        if self.training:
            with torch.no_grad():
                point_coords = get_uncertain_point_coords_with_randomness(
                    coarse_sem_seg_logits,
                    calculate_uncertainty,
                    self.train_num_points,
                    self.oversample_ratio,
                    self.importance_sample_ratio,
                )
            coarse_features = point_sample(coarse_sem_seg_logits, point_coords, align_corners=False)

            fine_grained_features = torch.cat(
                [
                    point_sample(in_feature, point_coords, align_corners=False)
                    for in_feature in backbone_features
                ],
                dim=1,
            )
            
            point_logits = self.point_head(fine_grained_features, coarse_features)
            
            return point_logits, point_coords
        else:
            sem_seg_logits = coarse_sem_seg_logits.clone()
            for _ in range(self.subdivision_steps):
                sem_seg_logits = F.interpolate(
                    sem_seg_logits, scale_factor=2, mode="bilinear", align_corners=False
                )
                uncertainty_map = calculate_uncertainty(sem_seg_logits)
                point_indices, point_coords = get_uncertain_point_coords_on_grid(
                    uncertainty_map, self.subdivision_num_points
                )
                
                fine_grained_features = torch.cat(
                [
                    point_sample(in_feature, point_coords, align_corners=False)
                    for in_feature in backbone_features
                ],
                dim=1,
                )

                coarse_features = point_sample(
                    coarse_sem_seg_logits, point_coords, align_corners=False
                )
                point_logits = self.point_head(fine_grained_features, coarse_features)

                # put sem seg point predictions to the right places on the upsampled grid.
                N, C, H, W = sem_seg_logits.shape
                point_indices = point_indices.unsqueeze(1).expand(-1, C, -1)
                sem_seg_logits = (
                    sem_seg_logits.reshape(N, C, H * W)
                    .scatter_(2, point_indices, point_logits)
                    .view(N, C, H, W)
                )
            return sem_seg_logits


class StandardPointHead(nn.Module):
    """
    A point head multi-layer perceptron which we model with conv1d layers with kernel 1. The head
    takes both fine-grained and coarse prediction features as its input.
    """

    def __init__(self, num_classes, input_channels, coarse_pred_each_layer=False):
        """
        The following attributes are parsed from config:
            fc_dim: the output dimension of each FC layers
            num_fc: the number of FC layers
            coarse_pred_each_layer: if True, coarse prediction features are concatenated to each
                layer's input
        """
        super(StandardPointHead, self).__init__()
        fc_dim                      = 256
        num_fc                      = 3
        cls_agnostic_mask           = False
        self.coarse_pred_each_layer = coarse_pred_each_layer

        fc_dim_in = input_channels + num_classes
        self.fc_layers = []
        for k in range(num_fc):
            fc = nn.Conv1d(fc_dim_in, fc_dim, kernel_size=1, stride=1, padding=0, bias=True)
            self.add_module("fc{}".format(k + 1), fc)
            self.fc_layers.append(fc)
            fc_dim_in = fc_dim
            fc_dim_in += num_classes if self.coarse_pred_each_layer else 0

        num_mask_classes = 1 if cls_agnostic_mask else num_classes
        self.predictor = nn.Conv1d(fc_dim_in, num_mask_classes, kernel_size=1, stride=1, padding=0)

        for layer in self.fc_layers:
            c2_msra_fill(layer)
        # use normal distribution initialization for mask prediction layer
        nn.init.normal_(self.predictor.weight, std=0.001)
        if self.predictor.bias is not None:
            nn.init.constant_(self.predictor.bias, 0)

    def forward(self, fine_grained_features, coarse_features):
        x = torch.cat((fine_grained_features, coarse_features), dim=1)
        for layer in self.fc_layers:
            x = F.relu(layer(x))
            if self.coarse_pred_each_layer:
                x = torch.cat((x, coarse_features), dim=1)
        return self.predictor(x)



import os
import sys
import json
import uuid
from arcgis.gis import GIS
from arcgis.mapping import create_symbol
from arcgis._impl.common._isd import InsensitiveDict
from ._base import BaseOGC
###########################################################################
class GeoRSSLayer(BaseOGC):
    """
    The GeoRSSLayer class is used to create a layer based on GeoRSS. GeoRSS is a
    way to add geographic information to an RSS feed. The GeoRSSLayer supports
    both GeoRSS-Simple and GeoRSS GML encodings, and multiple geometry types.

    It exports custom RSS tags as additional attribute fields in the form of
    simple strings or an array of JSON objects.

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Required string. The web address of the GeoRSS sevice.
    ---------------     --------------------------------------------------------------------
    copyright           Optional String. Describes limitations and usage of the data.
    ---------------     --------------------------------------------------------------------
    line_symbol         Optionl Dict. The symbol for the polyline data in the GeoRSS.
    ---------------     --------------------------------------------------------------------
    opacity             Optional Float.  This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.
    ---------------     --------------------------------------------------------------------
    point_symbol        Optionl Dict. The symbol for the point data in the GeoRSS.
    ---------------     --------------------------------------------------------------------
    polygon_symbol      Optionl Dict. The symbol for the polygon data in the GeoRSS.
    ---------------     --------------------------------------------------------------------
    title               Optional String. The title of the layer used to identify it in places such as the Legend and LayerList widgets.
    ---------------     --------------------------------------------------------------------
    scale               Optional Tuple. The min/max scale of the layer where the positions are: (min, max) as float values.
    ===============     ====================================================================

    """
    _line_symbol = None
    _point_symbol = None
    _polygon_symbol = None
    _type = "GeoRSS"
    #----------------------------------------------------------------------
    def __init__(self, url, **kwargs):
        super(GeoRSSLayer, self)
        self._id = kwargs.pop("id", uuid.uuid4().hex)
        self._url = url
        scale = kwargs.pop("scale", (0,0))
        assert isinstance(scale, (list, tuple)) and len(scale) == 2
        self._min_scale = scale[0]
        self._max_scale = scale[1]
        self._opacity = kwargs.pop('opacity', 1)
        self._copyright = kwargs.pop('copyright', None)
        self._title = kwargs.pop('title', None)
        self.point_symbol = kwargs.pop('point_symbol', None)
        self.polygon_symbol = kwargs.pop('polygon_symbol', None)
        self.line_symbol = kwargs.pop('line_symbol', None)
        self.title = kwargs.pop("title", "GeoRSS Feed")
    #----------------------------------------------------------------------
    @property
    def point_symbol(self) -> InsensitiveDict:
        """
        Gets/Sets the Point Symbol for Point Geometries
        
        :returns: InsensitiveDict
        """
        if self._point_symbol is None:
            self._point_symbol = InsensitiveDict(dict(create_symbol(geometry_type="point")))
        return self._point_symbol
    #----------------------------------------------------------------------
    @point_symbol.setter
    def point_symbol(self, value:dict):
        """
        Gets/Sets the Point Symbol for Point Geometries
        
        :returns: InsensitiveDict
        """        
        if isinstance(value, dict):
            self._point_symbol = InsensitiveDict(value)
        elif value is None:
            self._point_symbol = InsensitiveDict(dict(create_symbol(geometry_type="point")))
        else:
            self._point_symbol = InsensitiveDict(value)
    #----------------------------------------------------------------------
    @property
    def line_symbol(self) -> InsensitiveDict:
        """
        Gets/Sets the Line Symbol for Polyline Geometries
        
        :returns: InsensitiveDict
        """        
        if self._line_symbol is None:
            self._line_symbol = InsensitiveDict(dict(create_symbol(geometry_type="polyline")))
        return self._line_symbol
    #----------------------------------------------------------------------
    @line_symbol.setter
    def line_symbol(self, value:dict):
        """
        Gets/Sets the Line Symbol for Polyline Geometries
        
        :returns: InsensitiveDict
        """                
        if isinstance(value, dict):
            self._line_symbol = InsensitiveDict(value)
        elif value is None:
            self._line_symbol = InsensitiveDict(dict(create_symbol(geometry_type="polyline")))
        else:
            self._line_symbol = InsensitiveDict(value)
    #----------------------------------------------------------------------
    @property
    def polygon_symbol(self) -> InsensitiveDict:
        """
        Gets/Sets the Polygon Symbol for Polygon Geometries
        
        :returns: InsensitiveDict
        """                
        if self._polygon_symbol is None:
            self._polygon_symbol = InsensitiveDict(dict(create_symbol(geometry_type="polygon")))
        return self._polygon_symbol
    #----------------------------------------------------------------------
    @polygon_symbol.setter
    def polygon_symbol(self, value:dict):
        """
        Gets/Sets the Polygon Symbol for Polygon Geometries
        
        :returns: InsensitiveDict
        """                        
        if isinstance(value, dict):
            self._polygon_symbol = InsensitiveDict(value)
        elif value is None:
            self._polygon_symbol = InsensitiveDict(dict(create_symbol(geometry_type="polygon")))
        else:
            self._polygon_symbol = InsensitiveDict(value)
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self) -> dict:
        """Represents the MapView's widget JSON format"""
        add_layer =  {
            "type" : self._type,
            'url' : self._url,
            'opacity' : self.opacity,
            'minScale' : self.scale[0],
            'maxScale' : self.scale[1],
            'pointSymbol' : self.point_symbol._json(),
            'polygonSymbol' :self.polygon_symbol._json(),
            'lineSymbol' : self.line_symbol._json(),
            'id' : self._id,
            'title' : self.title
        }
        return add_layer

    @property
    def _operational_layer_json(self) -> dict:
        """Represents the WebMap's JSON format"""
        return self._lyr_json


import json
import os
from string import digits
from functools import lru_cache

from re import search
import time
import concurrent.futures
from arcgis._impl.common import _utils
from arcgis._impl.common._filters import StatisticFilter, TimeFilter, GeometryFilter
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._utils import _date_handler, chunks

from arcgis.features.feature import Feature, FeatureSet
from arcgis.geometry import SpatialReference
from arcgis.gis import Layer, _GISResource
from arcgis.mapping import MapImageLayer

###########################################################################
class MapFeatureLayer(Layer):
    """
    Map Feature Layers can be added to and visualized using maps.

    Map Feature Layers are created by publishing feature data to a GIS, and are exposed as a broader resource (Item) in the
    GIS. `MapFeatureLayer` objects can be obtained through the layers attribute on map image service Items in the GIS.
    """
    _metadatamanager = None
    _renderer = None
    _storage = None
    _dynamic_layer = None
    _attachments = None
    _time_filter = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None, container=None, dynamic_layer=None):
        """
        Constructs a map feature layer given a feature layer URL
        :param url: layer url
        :param gis: optional, the GIS that this layer belongs to. Required for secure map feature layers.
        :param container: optional, the MapImageLayer to which this layer belongs
        :param dynamic_layer: optional dictionary. If the layer is given a dynamic layer definition, this will be added to functions.
        """
        if gis is None:
            import arcgis
            gis = arcgis.env.active_gis
        if str(url).lower().endswith("/"):
            url = url[:-1]
        super(MapFeatureLayer, self).__init__(url, gis)

        self._attachments = None
        self._dynamic_layer = dynamic_layer
        self._time_filter = None
    #----------------------------------------------------------------------
    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict =  { 'type' : "FeatureLayer",
                      'url' : url }
        if self._token is not None:
            lyr_dict['serviceToken'] = self._token

        if self.filter is not None:
            lyr_dict['filter'] = self.filter
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += '?token=' + self._token

        lyr_dict = {'type': "FeatureLayer", 'url': url}

        if self.filter is not None:
            lyr_dict['options'] = json.dumps({ "definition_expression": self.filter })
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
    #----------------------------------------------------------------------
    @property
    @lru_cache(maxsize=10)
    def attachements(self):
        """
        Provides a manager to work with attachments if the MapFeatureLayer supports this functionality
        """
        if "supportsQueryAttachments" in self.properties and \
           self.properties["supportsQueryAttachments"] and \
           self._attachments is None:
            from arcgis.features.managers import AttachmentManager
            self._attachments = AttachmentManager(self)
        return self._attachments

    #----------------------------------------------------------------------
    @property
    def time_filter(self):
        """
        Starting at Enterprise 10.7.1+, instead of querying time-enabled map
        service layers or time-enabled feature service layers, a time filter
        can be specified. Time can be filtered as a single instant or by
        separating the two ends of a time extent with a comma.

        ================     =================================================
        **Input**            **Description**
        ----------------     -------------------------------------------------
        value                Required Datetime/List Datetime. This is a single
                             or list of start/stop date.
        ================     =================================================

        :returns: String of datetime values as milliseconds from epoch
        """
        return self._time_filter

    #----------------------------------------------------------------------
    @time_filter.setter
    def time_filter(self, value):
        """
        Starting at Enterprise 10.7.1+, instead of querying time-enabled map
        service layers or time-enabled feature service layers, a time filter
        can be specified. Time can be filtered as a single instant or by
        separating the two ends of a time extent with a comma.

        ================     =================================================
        **Input**            **Description**
        ----------------     -------------------------------------------------
        value                Required Datetime/List Datetime. This is a single
                             or list of start/stop date.
        ================     =================================================

        :returns: String of datetime values as milliseconds from epoch
        """
        import datetime as _dt
        v = []
        if isinstance(value, _dt.datetime):
            self._time_filter = f"{int(value.timestamp() * 1000)}" # means single time
        elif isinstance(value, (tuple, list)):
            for idx, d in enumerate(value):
                if idx > 1:
                    break
                if isinstance(d, _dt.datetime):
                    v.append(f"{int(value.timestamp() * 1000)}")
                elif isinstance(d, str):
                    v.append(d)
                elif d is None:
                    v.append("null")
            self._time_filter = ",".join(v)
        elif isinstance(value, str):
            self._time_filter = value
        elif value is None:
            self._time_filter = None
        else:
            raise Exception("Invalid datetime filter")

    #----------------------------------------------------------------------
    @property
    def renderer(self):
        """
        Get/Set the Renderer of the Map Feature Layer.  This overrides the default symbology when displaying it on a webmap.

        :returns: InsensitiveDict

        """
        from arcgis._impl.common._isd import InsensitiveDict
        if self._renderer is None and "drawingInfo" in self.properties:
            self._renderer = InsensitiveDict(dict(self.properties.drawingInfo.renderer))
        return self._renderer

    #----------------------------------------------------------------------
    @renderer.setter
    def renderer(self, value):
        """
        Get/Set the Renderer of the Map Feature Layer.  This overrides the default symbology when displaying it on a webmap.

        :returns: InsensitiveDict

        """
        from arcgis._impl.common._isd import InsensitiveDict
        if isinstance(value, (dict, PropertyMap)):
            self._renderer = InsensitiveDict(dict(value))
        elif value is None:
            self._renderer = None
        elif not isinstance(value, InsensitiveDict):
            raise ValueError("Invalid renderer type.")
        self._refresh = value

    #----------------------------------------------------------------------
    @classmethod
    def fromitem(cls, item, layer_id=0):
        """
        Creates a map feature layer from a GIS Item.
        The type of item should be a 'Map Image Layer'.
        The layer_id is the id of the layer in Map Service's layers.
        """
        from arcgis.mapping import MapImageLayer

        return MapImageLayer.fromitem(item).layers[layer_id]

    #----------------------------------------------------------------------
    @property
    def container(self):
        """
        The `MapImageLayer` to which this layer belongs.
        """
        if self._storage is None:
            self._storage = MapImageLayer(url=self._url.rstrip(digits)[:-1], gis=self._gis)
        return self._storage
    #----------------------------------------------------------------------
    def export_attachments(self, output_folder, label_field=None):
        """
        Exports attachments from the map feature layer in Imagenet format using the output_label_field.

        ====================================     ====================================================================
        **Argument**                             **Description**
        ------------------------------------     --------------------------------------------------------------------
        output_folder                            Required. Output folder where the attachments will be stored.
        ------------------------------------     --------------------------------------------------------------------
        label_field                              Optional. Field which contains the label/category of each feature.
                                                 If None, a default folder is created.
        ====================================     ====================================================================

        """
        import pandas
        import urllib
        import hashlib

        if not self.properties['hasAttachments']:
            raise Exception("Map Feature Layer doesn't have any attachments.")

        if not os.path.exists(output_folder):
            raise Exception("Invalid output folder path.")

        object_attachments_mapping = {}

        object_id_field = self.properties['objectIdField']

        dataframe_merged = pandas.merge(
            self.query().sdf,
            self.attachments.search(as_df=True),
            left_on=object_id_field,
            right_on='PARENTOBJECTID'
        )

        token = self._con.token

        internal_folder = os.path.join(output_folder, 'images')
        if not os.path.exists(internal_folder):
            os.mkdir(internal_folder)

        folder = 'images'
        for row in dataframe_merged.iterrows():

            if label_field is not None:
                folder = row[1][label_field]

            path = os.path.join(internal_folder, folder)

            if not os.path.exists(path):
                os.mkdir(path)

            if token is not None:
                url = '{}/{}/attachments/{}?token={}'.format(self.url, row[1][object_id_field], row[1]["ID"],
                                                             self._con.token)
            else:
                url = '{}/{}/attachments/{}'.format(self.url, row[1][object_id_field], row[1]["ID"])

            if not object_attachments_mapping.get(row[1][object_id_field]):
                object_attachments_mapping[row[1][object_id_field]] = []

            content = urllib.request.urlopen(url).read()

            md5_hash = hashlib.md5(content).hexdigest()
            attachment_path = os.path.join(path, f'{md5_hash}.jpg')

            object_attachments_mapping[row[1][object_id_field]].append(os.path.join('images', os.path.join(folder, f'{md5_hash}.jpg')))

            if os.path.exists(attachment_path):
                continue
            file = open(attachment_path, 'wb')
            file.write(content)
            file.close()

        mapping_path = os.path.join(output_folder, 'mapping.txt')
        file = open(mapping_path, 'w')
        file.write(json.dumps(object_attachments_mapping))
        file.close()

    #----------------------------------------------------------------------
    def generate_renderer(self, definition, where=None):
        """
        This operation groups data using the supplied definition
        (classification definition) and an optional where clause. The
        result is a renderer object. Use baseSymbol and colorRamp to define
        the symbols assigned to each class. If the operation is performed
        on a table, the result is a renderer object containing the data
        classes and no symbols.

        =================     ====================================================================
        **Argument**          **Description**
        -----------------     --------------------------------------------------------------------
        definition            required dict. The definition using the renderer that is generated.
                              Use either class breaks or unique value classification definitions.
                              See: https://resources.arcgis.com/en/help/rest/apiref/ms_classification.html
        -----------------     --------------------------------------------------------------------
        where                 optional string. A where clause for which the data needs to be
                              classified. Any legal SQL where clause operating on the fields in
                              the dynamic layer/table is allowed.
        =================     ====================================================================

        :returns: dictionary

        """
        if self._dynamic_layer:
            url = "%s/generateRenderer" % self._url.split('?')[0]
        else:
            url = "%s/generateRenderer" % self._url
        params = {'f' : 'json',
                  'classificationDef' : definition
                  }
        if where:
            params['where'] = where
        if self._dynamic_layer is not None:
            params['layer'] = self._dynamic_layer
        return self._con.post(path=url, postdata=params)

    #----------------------------------------------------------------------
    def _add_attachment(self, oid, file_path):
        """
        Adds an attachment to a feature service

        =================     ====================================================================
        **Argument**          **Description**
        -----------------     --------------------------------------------------------------------
        oid                   Required string/integer. OBJECTID value to add attachment to.
        -----------------     --------------------------------------------------------------------
        file_path             Required string. Location of the file to attach.
        =================     ====================================================================

        :returns: dictionary

        """
        if (os.path.getsize(file_path) >> 20) <= 9:
            params = {'f': 'json'}
            if self._dynamic_layer:
                attach_url = self._url.split('?')[0] + "/%s/addAttachment" % oid
                params['layer'] = self._dynamic_layer
            else:
                attach_url = self._url + "/%s/addAttachment" % oid
            files = {'attachment': file_path}
            res = self._con.post(path=attach_url,
                                 postdata=params,
                                 files=files, token=self._token)
            return res
        else:
            params = {'f': 'json'}
            container = self.container
            itemid = container.upload(file_path)
            if self._dynamic_layer:
                attach_url = self._url.split('?')[0] + "/%s/addAttachment" % oid
                params['layer'] = self._dynamic_layer
            else:
                attach_url = self._url + "/%s/addAttachment" % oid
            params['uploadId'] = itemid
            res = self._con.post(attach_url,
                                 params)
            if res['addAttachmentResult']['success'] == True:
                container._delete_upload(itemid)
            return res
    # ----------------------------------------------------------------------
    def _delete_attachment(self, oid, attachment_id):
        """
        Removes an attachment from a feature service feature

        =================     ====================================================================
        **Argument**          **Description**
        -----------------     --------------------------------------------------------------------
        oid                   Required string/integer. OBJECTID value to add attachment to.
        -----------------     --------------------------------------------------------------------
        attachment_id         Required integer. Id of the attachment to erase.
        =================     ====================================================================

        :returns: dictionary
        """
        params = {
            "f": "json",
            "attachmentIds": "%s" % attachment_id
        }
        if self._dynamic_layer:
            url = self._url.split('?')[0] + "/%s/deleteAttachments" % oid
            params['layer'] = self._dynamic_layer
        else:
            url = self._url + "/%s/deleteAttachments" % oid
        return self._con.post(url, params, token=self._token)

    # ----------------------------------------------------------------------
    def _update_attachment(self, oid, attachment_id, file_path):
        """
        Updates an existing attachment with a new file

        =================     ====================================================================
        **Argument**          **Description**
        -----------------     --------------------------------------------------------------------
        oid                   Required string/integer. OBJECTID value to add attachment to.
        -----------------     --------------------------------------------------------------------
        attachment_id         Required integer. Id of the attachment to erase.
        -----------------     --------------------------------------------------------------------
        file_path             Required string. Path to new attachment
        =================     ====================================================================

        :returns: dictionary

        """
        params = {
            "f": "json",
            "attachmentId": "%s" % attachment_id
        }
        files = {'attachment': file_path}
        if self._dynamic_layer is not None:
            url = self.url.split('?')[0] + f"/{oid}/updateAttachment"
            params['layer'] = self._dynamic_layer
        else:
            url = self._url + f"/{oid}/updateAttachment"
        res = self._con.post(path=url,
                             postdata=params,
                             files=files, token=self._token)
        return res

    # ----------------------------------------------------------------------
    def _list_attachments(self, oid):
        """ list attachments for a given OBJECT ID """

        params = {
            "f": "json"
        }
        if self._dynamic_layer is not None:
            url = self.url.split('?')[0] + "/%s/attachments" % oid
            params['layer'] = self._dynamic_layer
        else:
            url = self._url + "/%s/attachments" % oid
        return self._con.get(path=url, params=params, token=self._token)

    # ----------------------------------------------------------------------
    def get_unique_values(self, attribute, query_string='1=1'):
        """Return a list of unique values for a given attribute

        ===============================     ====================================================================
        **Argument**                        **Description**
        -------------------------------     --------------------------------------------------------------------
        attribute                           Required string. The map feature layer attribute to query.
        -------------------------------     --------------------------------------------------------------------
        query_string                        Optional string. SQL Query that will be used to filter attributes
                                            before unique values are returned.
                                            ex. "name_2 like '%K%'"
        ===============================     ====================================================================
        """

        result = self.query(query_string, return_geometry=False, out_fields=attribute,return_distinct_values=True)
        return [feature.attributes[attribute] for feature in result.features]
    # ----------------------------------------------------------------------
    def query(self,
              where="1=1",
              text=None, # new
              out_fields="*",
              time_filter=None,
              geometry_filter=None,
              return_geometry=True,
              return_count_only=False,
              return_ids_only=False,
              return_distinct_values=False,
              return_extent_only=False,
              group_by_fields_for_statistics=None,
              statistic_filter=None,
              result_offset=None,
              result_record_count=None,
              object_ids=None,
              distance=None,
              units=None,
              max_allowable_offset=None,
              out_sr=None,
              geometry_precision=None,
              gdb_version=None,
              order_by_fields=None,
              out_statistics=None,
              return_z=False,
              return_m=False,
              multipatch_option=None,
              quantization_parameters=None,
              return_centroid=False,
              return_all_records=True,
              result_type=None,
              historic_moment=None,
              sql_format=None,
              return_true_curves=False,
              return_exceeded_limit_features=None,
              as_df=False,
              datum_transformation=None,
              range_values=None,
              parameter_values=None,
              **kwargs):
        """
        Queries a map feature layer based on a sql statement

        ===============================     ====================================================================
        **Argument**                        **Description**
        -------------------------------     --------------------------------------------------------------------
        where                               Optional string. The default is 1=1. The selection sql statement.
        -------------------------------     --------------------------------------------------------------------
        text                                Optional String. A literal search text. If the layer has a display
                                            field associated with it, the server searches for this text in this
                                            field.
        -------------------------------     --------------------------------------------------------------------
        out_fields                          Optional List of field names to return. Field names can be specified
                                            either as a List of field names or as a comma separated string.
                                            The default is "*", which returns all the fields.
        -------------------------------     --------------------------------------------------------------------
        object_ids                          Optional string. The object IDs of this layer or table to be queried.
                                            The object ID values should be a comma-separated string.
        -------------------------------     --------------------------------------------------------------------
        distance                            Optional integer. The buffer distance for the input geometries.
                                            The distance unit is specified by units. For example, if the
                                            distance is 100, the query geometry is a point, units is set to
                                            meters, and all points within 100 meters of the point are returned.
        -------------------------------     --------------------------------------------------------------------
        units                               Optional string. The unit for calculating the buffer distance. If
                                            unit is not specified, the unit is derived from the geometry spatial
                                            reference. If the geometry spatial reference is not specified, the
                                            unit is derived from the feature service data spatial reference.
                                            This parameter only applies if supportsQueryWithDistance is true.
                                            Values: esriSRUnit_Meter | esriSRUnit_StatuteMile |
                                                    esriSRUnit_Foot | esriSRUnit_Kilometer |
                                                    esriSRUnit_NauticalMile | esriSRUnit_USNauticalMile
        -------------------------------     --------------------------------------------------------------------
        time_filter                         Optional list. The format is of [<startTime>, <endTime>] using
                                            datetime.date, datetime.datetime or timestamp in milliseconds.
                                            Syntax: time_filter=[<startTime>, <endTime>] ; specified as
                                                    datetime.date, datetime.datetime or timestamp in
                                                    milliseconds
        -------------------------------     --------------------------------------------------------------------
        geometry_filter                     Optional from arcgis.geometry.filter. Allows for the information to
                                            be filtered on spatial relationship with another geometry.
        -------------------------------     --------------------------------------------------------------------
        max_allowable_offset                Optional float. This option can be used to specify the
                                            max_allowable_offset to be used for generalizing geometries returned
                                            by the query operation.
                                            The max_allowable_offset is in the units of out_sr. If out_sr is not
                                            specified, max_allowable_offset is assumed to be in the unit of the
                                            spatial reference of the layer.
        -------------------------------     --------------------------------------------------------------------
        out_sr                              Optional Integer. The WKID for the spatial reference of the returned
                                            geometry.
        -------------------------------     --------------------------------------------------------------------
        geometry_precision                  Optional Integer. This option can be used to specify the number of
                                            decimal places in the response geometries returned by the query
                                            operation.
                                            This applies to X and Y values only (not m or z-values).
        -------------------------------     --------------------------------------------------------------------
        gdb_version                         Optional string. The geodatabase version to query. This parameter
                                            applies only if the isDataVersioned property of the layer is true.
                                            If this is not specified, the query will apply to the published
                                            map's version.
        -------------------------------     --------------------------------------------------------------------
        return_geometry                     Optional boolean. If true, geometry is returned with the query.
                                            Default is true.
        -------------------------------     --------------------------------------------------------------------
        return_distinct_values              Optional boolean.  If true, it returns distinct values based on the
                                            fields specified in out_fields. This parameter applies only if the
                                            supportsAdvancedQueries property of the layer is true.
        -------------------------------     --------------------------------------------------------------------
        return_ids_only                     Optional boolean. Default is False.  If true, the response only
                                            includes an array of object IDs. Otherwise, the response is a
                                            feature set.
        -------------------------------     --------------------------------------------------------------------
        return_count_only                   Optional boolean. If true, the response only includes the count
                                            (number of features/records) that would be returned by a query.
                                            Otherwise, the response is a feature set. The default is false. This
                                            option supersedes the returnIdsOnly parameter. If
                                            returnCountOnly = true, the response will return both the count and
                                            the extent.
        -------------------------------     --------------------------------------------------------------------
        return_extent_only                  Optional boolean. If true, the response only includes the extent of
                                            the features that would be returned by the query. If
                                            returnCountOnly=true, the response will return both the count and
                                            the extent.
                                            The default is false. This parameter applies only if the
                                            supportsReturningQueryExtent property of the layer is true.
        -------------------------------     --------------------------------------------------------------------
        order_by_fields                     Optional string. One or more field names on which the
                                            features/records need to be ordered. Use ASC or DESC for ascending
                                            or descending, respectively, following every field to control the
                                            ordering.
                                            example: STATE_NAME ASC, RACE DESC, GENDER
        -------------------------------     --------------------------------------------------------------------
        group_by_fields_for_statistics      Optional string. One or more field names on which the values need to
                                            be grouped for calculating the statistics.
                                            example: STATE_NAME, GENDER
        -------------------------------     --------------------------------------------------------------------
        out_statistics                      Optional string. The definitions for one or more field-based
                                            statistics to be calculated.

                                            Syntax:

                                            [
                                                {
                                                  "statisticType": "<count | sum | min | max | avg | stddev | var>",
                                                  "onStatisticField": "Field1",
                                                  "outStatisticFieldName": "Out_Field_Name1"
                                                },
                                                {
                                                  "statisticType": "<count | sum | min | max | avg | stddev | var>",
                                                  "onStatisticField": "Field2",
                                                  "outStatisticFieldName": "Out_Field_Name2"
                                                }
                                            ]
        -------------------------------     --------------------------------------------------------------------
        return_z                            Optional boolean. If true, Z values are included in the results if
                                            the features have Z values. Otherwise, Z values are not returned.
                                            The default is False.
        -------------------------------     --------------------------------------------------------------------
        return_m                            Optional boolean. If true, M values are included in the results if
                                            the features have M values. Otherwise, M values are not returned.
                                            The default is false.
        -------------------------------     --------------------------------------------------------------------
        multipatch_option                   Optional x/y footprint. This option dictates how the geometry of
                                            a multipatch feature will be returned.
        -------------------------------     --------------------------------------------------------------------
        result_offset                       Optional integer. This option can be used for fetching query results
                                            by skipping the specified number of records and starting from the
                                            next record (that is, resultOffset + 1th). This option is ignored
                                            if return_all_records is True (i.e. by default).
        -------------------------------     --------------------------------------------------------------------
        result_record_count                 Optional integer. This option can be used for fetching query results
                                            up to the result_record_count specified. When result_offset is
                                            specified but this parameter is not, the map service defaults it to
                                            max_record_count. The maximum value for this parameter is the value
                                            of the layer's max_record_count property. This option is ignored if
                                            return_all_records is True (i.e. by default).
        -------------------------------     --------------------------------------------------------------------
        quantization_parameters             Optional dict. Used to project the geometry onto a virtual grid,
                                            likely representing pixels on the screen.
        -------------------------------     --------------------------------------------------------------------
        return_centroid                     Optional boolean. Used to return the geometry centroid associated
                                            with each feature returned. If true, the result includes the geometry
                                            centroid. The default is false.
        -------------------------------     --------------------------------------------------------------------
        return_all_records                  Optional boolean. When True, the query operation will call the
                                            service until all records that satisfy the where_clause are
                                            returned. Note: result_offset and result_record_count will be
                                            ignored if return_all_records is True. Also, if return_count_only,
                                            return_ids_only, or return_extent_only are True, this parameter
                                            will be ignored.
        -------------------------------     --------------------------------------------------------------------
        result_type                         Optional string. The result_type parameter can be used to control
                                            the number of features returned by the query operation.
                                            Values: None | standard | tile
        -------------------------------     --------------------------------------------------------------------
        historic_moment                     Optional integer. The historic moment to query. This parameter
                                            applies only if the layer is archiving enabled and the
                                            supportsQueryWithHistoricMoment property is set to true. This
                                            property is provided in the layer resource.

                                            If historic_moment is not specified, the query will apply to the
                                            current features.
        -------------------------------     --------------------------------------------------------------------
        sql_format                          Optional string.  The sql_format parameter can be either standard
                                            SQL92 standard or it can use the native SQL of the underlying
                                            datastore native. The default is none which means the sql_format
                                            depends on useStandardizedQuery parameter.
                                            Values: none | standard | native
        -------------------------------     --------------------------------------------------------------------
        return_true_curves                  Optional boolean. When set to true, returns true curves in output
                                            geometries. When set to false, curves are converted to densified
                                            polylines or polygons.
        -------------------------------     --------------------------------------------------------------------
        return_exceeded_limit_features      Optional boolean. Optional parameter which is true by default. When
                                            set to true, features are returned even when the results include
                                            'exceededTransferLimit': True.

                                            When set to false and querying with resultType = tile features are
                                            not returned when the results include 'exceededTransferLimit': True.
                                            This allows a client to find the resolution in which the transfer
                                            limit is no longer exceeded without making multiple calls.
        -------------------------------     --------------------------------------------------------------------
        as_df                               Optional boolean.  If True, the results are returned as a DataFrame
                                            instead of a FeatureSet.
        -------------------------------     --------------------------------------------------------------------
        datum_transformation                Optional Integer/Dictionary.  This parameter applies a datum transformation while
                                            projecting geometries in the results when out_sr is different than the layer's spatial
                                            reference. When specifying transformations, you need to think about which datum
                                            transformation best projects the layer (not the feature service) to the `outSR` and
                                            `sourceSpatialReference` property in the layer properties. For a list of valid datum
                                            transformation ID values ad well-known text strings, see `Coordinate systems and
                                            transformations <https://developers.arcgis.com/net/latest/wpf/guide/coordinate-systems-and-transformations.htm>`_.
                                            For more information on datum transformations, please see the transformation
                                            parameter in the `Project operation <https://developers.arcgis.com/rest/services-reference/project.htm>`_.

                                            **Examples**


                                                ===========     ===================================
                                                Inputs          Description
                                                -----------     -----------------------------------
                                                WKID            Integer. Ex: datum_transformation=4326
                                                -----------     -----------------------------------
                                                WKT             Dict. Ex: datum_transformation={"wkt": "<WKT>"}
                                                -----------     -----------------------------------
                                                Composite       Dict. Ex: datum_transformation=```{'geoTransforms':[{'wkid':<id>,'forward':<true|false>},{'wkt':'<WKT>','forward':<True|False>}]}```
                                                ===========     ===================================
        -------------------------------     --------------------------------------------------------------------
        range_values                        Optional List. Allows you to filter features from the layer that are
                                            within the specified range instant or extent.

                                            :Syntax:

                                            [
                                                {
                                                  "name": "range name",
                                                  "value": <value> or [ <value1>, <value2> ]  # single value or a value-range
                                                      # None is allowed in value-range case -- that means infinity
                                                      # e.g. [None, 1500] means all features with values <= 1500
                                                      # [1000, None] means all features with values >= 1000
                                                  },
                                                  {
                                                    "name": "range name 2",
                                                    "value": <value> or  [ <value3>, <value4> ]
                                                  }
                                                }
                                            ]

        -------------------------------     --------------------------------------------------------------------
        parameter_values                    Optional Dict. Allows you to filter the features layers by specifying
                                            value(s) to an array of pre-authored parameterized filters for those
                                            layers. When value is not specified for any parameter in a request,
                                            the default value, that is assigned during authoring time, gets used
                                            instead.

                                            When a parameterInfo allows multiple values, you must pass them in an array.

                                            Note: Check parameterInfos at the layer resources for the available
                                            parameterized filters, their default values and expected data type.
        -------------------------------     --------------------------------------------------------------------
        kwargs                              Optional dict. Optional parameters that can be passed to the Query
                                            function.  This will allow users to pass additional parameters not
                                            explicitly implemented on the function. A complete list of functions
                                            available is documented on the Query REST API.
        ===============================     ====================================================================

        :returns: A FeatureSet containing the features matching the query unless another return type is specified, such as count
        """
        as_raw = as_df
        if self._dynamic_layer is None:
            url = self._url + "/query"
        else:
            url = "%s/query" % self._url.split('?')[0]

        params = {"f": "json"}
        if self._dynamic_layer is not None:
            params['layer'] = self._dynamic_layer
        if result_type is not None:
            params['resultType'] = result_type
        if historic_moment is not None:
            params['historicMoment'] = historic_moment
        if sql_format is not None:
            params['sqlFormat'] = sql_format
        if return_true_curves is not None:
            params['returnTrueCurves'] = return_true_curves
        if return_exceeded_limit_features is not None:
            params['returnExceededLimitFeatures'] = return_exceeded_limit_features
        params['where'] = where
        params['returnGeometry'] = return_geometry
        params['returnDistinctValues'] = return_distinct_values
        params['returnCentroid'] = return_centroid
        params['returnCountOnly'] = return_count_only
        params['returnExtentOnly'] = return_extent_only
        params['returnIdsOnly'] = return_ids_only
        params['returnZ'] = return_z
        params['returnM'] = return_m
        if parameter_values:
            params['parameterValues']
        if range_values:
            params['rangeValues'] = range_values
        if not datum_transformation is None:
            params['datumTransformation'] = datum_transformation

        # convert out_fields to a comma separated string
        if isinstance(out_fields, (list, tuple)):
            out_fields = ','.join(out_fields)

        if out_fields != '*' and not return_distinct_values:
            try:
                # Check if object id field is in out_fields.
                # If it isn't, add it
                object_id_field = [x.name for x in self.properties.fields if x.type == "esriFieldTypeOID"][0]
                if object_id_field not in out_fields.split(','):
                    out_fields = object_id_field + "," + out_fields
            except (IndexError, AttributeError):
                pass
        params['outFields'] = out_fields
        if return_count_only or return_extent_only or return_ids_only:
            return_all_records = False
        if result_record_count and not return_all_records:
            params['resultRecordCount'] = result_record_count
        if result_offset and not return_all_records:
            params['resultOffset'] = result_offset
        if quantization_parameters:
            params['quantizationParameters'] = quantization_parameters
        if multipatch_option:
            params['multipatchOption'] = multipatch_option
        if order_by_fields:
            params['orderByFields'] = order_by_fields
        if group_by_fields_for_statistics:
            params['groupByFieldsForStatistics'] = group_by_fields_for_statistics
        if statistic_filter and \
                isinstance(statistic_filter, StatisticFilter):
            params['outStatistics'] = statistic_filter.filter
        if out_statistics:
            params['outStatistics'] = out_statistics
        if text:
            params['text'] = text
        if out_sr:
            params['outSR'] = out_sr
        if max_allowable_offset:
            params['maxAllowableOffset'] = max_allowable_offset
        if gdb_version:
            params['gdbVersion'] = gdb_version
        if geometry_precision:
            params['geometryPrecision'] = geometry_precision
        if object_ids:
            params['objectIds'] = object_ids
        if distance:
            params['distance'] = distance
        if units:
            params['units'] = units

        if time_filter is None and self.time_filter:
            params['time'] = self.time_filter
        elif time_filter is not None:
            if type(time_filter) is list:
                starttime = _date_handler(time_filter[0])
                endtime = _date_handler(time_filter[1])
                if starttime is None:
                    starttime = 'null'
                if endtime is None:
                    endtime = 'null'
                params['time'] = "%s,%s" % (starttime, endtime)
            elif isinstance(time_filter, dict):
                for key, val in time_filter.items():
                    params[key] = val
            else:
                params['time'] = _date_handler(time_filter)

        if geometry_filter and \
                isinstance(geometry_filter, GeometryFilter):
            for key, val in geometry_filter.filter:
                params[key] = val
        elif geometry_filter and \
                isinstance(geometry_filter, dict):
            for key, val in geometry_filter.items():
                params[key] = val
        if len(kwargs) > 0:
            for key, val in kwargs.items():
                if key in ('returnCountOnly','returnExtentOnly','returnIdsOnly') and val:
                    # If these keys are passed in as kwargs instead of parameters, set return_all_records
                    return_all_records = False
                params[key] = val
                del key, val

        if not return_all_records or "outStatistics" in params:
            if as_df:
                return self._query_df(url, params)
            return self._query(url, params, raw=as_raw)

        params['returnCountOnly'] = True
        record_count = self._query(url, params, raw=as_raw)
        if 'maxRecordCount' in self.properties:
            max_records = self.properties['maxRecordCount']
        else:
            max_records = 1000

        supports_pagination = True
        if ('advancedQueryCapabilities' not in self.properties or \
                'supportsPagination' not in self.properties['advancedQueryCapabilities'] or \
                not self.properties['advancedQueryCapabilities']['supportsPagination']):
            supports_pagination = False

        params['returnCountOnly'] = False
        if record_count == 0 and as_df:
            from arcgis.features.geo._array import GeoArray
            import numpy as np
            import pandas as pd
            _fld_lu = {
                "esriFieldTypeSmallInteger" : np.int32,
                "esriFieldTypeInteger" : np.int64,
                "esriFieldTypeSingle" : np.int32,
                "esriFieldTypeDouble" : float,
                "esriFieldTypeFloat" : float,
                "esriFieldTypeString" : str,
                "esriFieldTypeDate" : np.datetime64,
                "esriFieldTypeOID" : np.int64,
                "esriFieldTypeGeometry" : object,
                "esriFieldTypeBlob" : object,
                "esriFieldTypeRaster" : object,
                "esriFieldTypeGUID" : str,
                "esriFieldTypeGlobalID" : str,
                "esriFieldTypeXML" : object
            }
            columns = {}
            for fld in self.properties.fields:
                fld = dict(fld)
                columns[fld['name']] = _fld_lu[fld['type']]
            if "geometryType" in self.properties and \
               not self.properties.geometryType is None:
                columns['SHAPE'] = object
            df = pd.DataFrame([], columns=columns.keys()).astype(columns, True)
            if 'SHAPE' in df.columns:
                df['SHAPE'] = GeoArray([])
                df.spatial.set_geometry("SHAPE")
                df.spatial.renderer = self.renderer
                df.spatial._meta.source = self
            return df
        elif record_count <= max_records:
            if supports_pagination and record_count > 0:
                params['resultRecordCount'] = record_count
            if as_df:
                import pandas as pd
                df = self._query_df(url, params)
                dt_fields = [fld['name'] for fld in self.properties.fields \
                             if fld['type'] == 'esriFieldTypeDate']
                if 'SHAPE' in df.columns:
                    df.spatial.set_geometry('SHAPE')
                    df.spatial.renderer = self.renderer
                    df.spatial._meta.source = self
                for fld in dt_fields:
                    try:
                        if fld in df.columns:
                            df[fld] = pd.to_datetime(df[fld]/1000,
                                                 infer_datetime_format=True,
                                                 unit='s')
                    except:
                        if fld in df.columns:
                            df[fld] = pd.to_datetime(df[fld], infer_datetime_format=True)
                return df

            return self._query(url, params, raw=as_raw)

        result = None
        i = 0
        count = 0
        df = None
        dfs = []
        if not supports_pagination:
            params['returnIdsOnly'] = True
            oid_info = self._query(url, params, raw=as_raw)
            params['returnIdsOnly'] = False
            for ids in chunks(oid_info['objectIds'], max_records):
                ids = [str(i) for i in ids]
                sql = "%s in (%s)" % (oid_info['objectIdFieldName'], ",".join(ids))
                params['where'] = sql
                if not as_df:
                    records = self._query(url, params, raw=as_raw)
                    if result:
                        if 'features' in result:
                            result['features'].append(records['features'])
                        else:
                            result.features.extend(records.features)
                    else:
                        result = records
                else:
                    df = self._query_df(url, params)
                    dfs.append(df)
        else:
            while True:
                params['resultRecordCount'] = max_records
                params['resultOffset'] = max_records * i
                if not as_df:
                    records = self._query(url, params, raw=as_raw)

                    if result:
                        if 'features' in result:
                            result['features'].append(records['features'])
                        else:
                            result.features.extend(records.features)
                    else:
                        result = records

                    if len(records.features) < max_records:
                        break
                else:

                    df = self._query_df(url, params)
                    count += len(df)
                    dfs.append(df)
                    if count == record_count:
                        break
                i += 1
        if as_df:
            import pandas as pd
            dt_fields = [fld['name'] for fld in self.properties.fields \
                         if fld['type'] == 'esriFieldTypeDate']
            if len(dfs) == 1:
                df = dfs[0]
            else:
                df = pd.concat(dfs, sort=True)
                df.reset_index(drop=True, inplace=True)
            if 'SHAPE' in df.columns:
                df.spatial.set_geometry('SHAPE')
                df.spatial.renderer = self.renderer
                df.spatial._meta.source = self
            for fld in dt_fields:
                if fld in df.columns:

                    try:
                        df[fld] = pd.to_datetime(df[fld]/1000,
                                                 infer_datetime_format=True,
                                                 unit='s')
                    except:
                        df[fld] = pd.to_datetime(df[fld], infer_datetime_format=True, errors='coerce')
            return df
        return result
    # ----------------------------------------------------------------------
    def query_related_records(self,
                              object_ids,
                              relationship_id,
                              out_fields="*",
                              definition_expression=None,
                              return_geometry=True,
                              max_allowable_offset=None,
                              geometry_precision=None,
                              out_wkid=None,
                              gdb_version=None,
                              return_z=False,
                              return_m=False,
                              historic_moment=None,
                              return_true_curve=False):
        """
        The Query operation is performed on a feature service layer
        resource. The result of this operation are feature sets grouped
        by source layer/table object IDs. Each feature set contains
        Feature objects including the values for the fields requested by
        the user. For related layers, if you request geometry
        information, the geometry of each feature is also returned in
        the feature set. For related tables, the feature set does not
        include geometries.

        ======================     ====================================================================
        **Argument**               **Description**
        ----------------------     --------------------------------------------------------------------
        object_ids                 Required string. The object IDs of the table/layer to be queried
        ----------------------     --------------------------------------------------------------------
        relationship_id            Required string. The ID of the relationship to be queried.
        ----------------------     --------------------------------------------------------------------
        out_fields                 Required string. the list of fields from the related table/layer
                                   to be included in the returned feature set. This list is a comma
                                   delimited list of field names. If you specify the shape field in the
                                   list of return fields, it is ignored. To request geometry, set
                                   return_geometry to true. You can also specify the wildcard "*" as
                                   the value of this parameter. In this case, the results will include
                                   all the field values.
        ----------------------     --------------------------------------------------------------------
        definition_expression      Optional string. The definition expression to be applied to the
                                   related table/layer. From the list of objectIds, only those records
                                   that conform to this expression are queried for related records.
        ----------------------     --------------------------------------------------------------------
        return_geometry            Optional boolean. If true, the feature set includes the geometry
                                   associated with each feature. The default is true.
        ----------------------     --------------------------------------------------------------------
        max_allowable_offset       Optional float. This option can be used to specify the
                                   max_allowable_offset to be used for generalizing geometries returned
                                   by the query operation. The max_allowable_offset is in the units of
                                   the outSR. If out_wkid is not specified, then max_allowable_offset
                                   is assumed to be in the unit of the spatial reference of the map.
        ----------------------     --------------------------------------------------------------------
        geometry_precision         Optional integer. This option can be used to specify the number of
                                   decimal places in the response geometries.
        ----------------------     --------------------------------------------------------------------
        out_wkid                   Optional Integer. The spatial reference of the returned geometry.
        ----------------------     --------------------------------------------------------------------
        gdb_version                Optional string. The geodatabase version to query. This parameter
                                   applies only if the isDataVersioned property of the layer queried is
                                   true.
        ----------------------     --------------------------------------------------------------------
        return_z                   Optional boolean. If true, Z values are included in the results if
                                   the features have Z values. Otherwise, Z values are not returned.
                                   The default is false.
        ----------------------     --------------------------------------------------------------------
        return_m                   Optional boolean. If true, M values are included in the results if
                                   the features have M values. Otherwise, M values are not returned.
                                   The default is false.
        ----------------------     --------------------------------------------------------------------
        historic_moment            Optional Integer/datetime. The historic moment to query. This parameter
                                   applies only if the supportsQueryWithHistoricMoment property of the
                                   layers being queried is set to true. This setting is provided in the
                                   layer resource.

                                   If historic_moment is not specified, the query will apply to the
                                   current features.

                                   Syntax: historic_moment=<Epoch time in milliseconds>
        ----------------------     --------------------------------------------------------------------
        return_true_curves         Optional boolean. Optional parameter that is false by default. When
                                   set to true, returns true curves in output geometries; otherwise,
                                   curves are converted to densified polylines or polygons.
        ======================     ====================================================================


        :return: dict


        """
        params = {
            "f": "json",
            "objectIds": object_ids,
            "relationshipId": relationship_id,
            "outFields": out_fields,
            "returnGeometry": return_geometry,
            "returnM": return_m,
            "returnZ": return_z
        }
        if historic_moment:
            if hasattr(historic_moment, "timestamp"):
                historic_moment = int(historic_moment.timestamp() * 1000)
            params['historicMoment'] = historic_moment
        if return_true_curve:
            params['returnTrueCurves'] = return_true_curve
        if self._dynamic_layer is not None:
            params['layer'] = self._dynamic_layer
        if gdb_version is not None:
            params['gdbVersion'] = gdb_version
        if definition_expression is not None:
            params['definitionExpression'] = definition_expression
        if out_wkid is not None and \
                isinstance(out_wkid, SpatialReference):
            params['outSR'] = out_wkid
        elif out_wkid is not None and \
                isinstance(out_wkid, dict):
            params['outSR'] = out_wkid
        if max_allowable_offset is not None:
            params['maxAllowableOffset'] = max_allowable_offset
        if geometry_precision is not None:
            params['geometryPrecision'] = geometry_precision
        if self._dynamic_layer is None:
            qrr_url = self._url + "/queryRelatedRecords"
        else:
            qrr_url = "%s/queryRelatedRecords" % self._url.split('?')[0]

        return self._con.post(path=qrr_url, postdata=params, token=self._token)

    # ----------------------------------------------------------------------
    def get_html_popup(self, oid):
        """
        The htmlPopup resource provides details about the HTML pop-up
        authored by the user using ArcGIS for Desktop.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        oid                 Optional string. Object id of the feature to get the HTML popup.
        ===============     ====================================================================


        :return: string

        """
        if self.properties.htmlPopupType != "esriServerHTMLPopupTypeNone":
            pop_url = self._url + "/%s/htmlPopup" % oid
            params = {
                'f': "json"
            }

            return self._con.get(path=pop_url, params=params, token=self._token)
        return ""
    # ----------------------------------------------------------------------
    def _status_via_url(self, con, url, params):
        """
        performs the asynchronous check to see if the operation finishes
        """
        status_allowed = ['Pending', 'InProgress', 'Completed', 'Failed ImportChanges',
                          'ExportChanges', 'ExportingData', 'ExportingSnapshot',
                          'ExportAttachments', 'ImportAttachments', 'ProvisioningReplica',
                          'UnRegisteringReplica', 'CompletedWithErrors']
        status = con.get(url, params)
        while status['status'] in status_allowed and \
              status['status'] != 'Completed':
            if status['status'] == 'Completed':
                return status
            elif status['status'] == 'CompletedWithErrors':
                break
            elif 'fail' in status['status'].lower():
                break
            elif 'error' in status['status'].lower():
                break
            status = con.get(url, params)
        return status

    # ----------------------------------------------------------------------
    def _query(self, url, params, raw=False):
        """ returns results of query """
        try:
            result = self._con.post(path=url,
                                    postdata=params, token=self._token)
        except Exception as queryException:
            error_list = ["Error performing query operation", "HTTP Error 504: GATEWAY_TIMEOUT"]
            if any(ele in queryException.__str__() for ele in error_list):
                # half the max record count
                max_record = int(params['resultRecordCount']) if 'resultRecordCount' in params else 1000
                offset = int(params['resultOffset']) if 'resultOffset' in params else 0
                # reduce this number to 125 if you still sees 500/504 error
                if max_record < 250:
                    # when max_record is lower than 250, but still getting error 500 or 504, just exit with exception
                    raise queryException
                else:
                    max_rec = int((max_record + 1) / 2)
                    i = 0
                    result = None
                    while max_rec * i < max_record:
                        params['resultRecordCount'] = max_rec if max_rec*(i+1) <= max_record else (max_record - max_rec*i)
                        params['resultOffset'] = offset + max_rec * i
                        try:
                            records = self._query(url, params, raw=True)
                            if result:
                                for feature in records['features']:
                                    result['features'].append(feature)
                            else:
                                result = records
                            i += 1
                        except Exception as queryException2:
                            raise queryException2

            else:
                raise queryException

        def is_true(x):
            if isinstance(x, bool) and x:
                return True
            elif isinstance(x, str) and x.lower() == 'true':
                return True
            else:
                return False

        if 'error' in result:
            raise ValueError(result)
        if 'returnCountOnly' in params and is_true(params['returnCountOnly']):
            return result['count']
        elif 'returnIdsOnly' in params and is_true(params['returnIdsOnly']):
            return result
        elif 'extent' in result:
            return result
        elif is_true(raw):
            return result
        else:
            return FeatureSet.from_dict(result)

    # ----------------------------------------------------------------------
    def _query_df(self, url, params):
        """ returns results of a query as a pd.DataFrame"""
        import pandas as pd
        from arcgis.features import GeoAccessor, GeoSeriesAccessor
        import numpy as np
        if [float(i) for i in pd.__version__.split('.')] < [1,0,0]:
            _fld_lu = {
                "esriFieldTypeSmallInteger" : np.int32,
                "esriFieldTypeInteger" : np.int64,
                "esriFieldTypeSingle" : np.int32,
                "esriFieldTypeDouble" : float,
                "esriFieldTypeFloat" : float,
                "esriFieldTypeString" : str,
                "esriFieldTypeDate" : pd.datetime,
                "esriFieldTypeOID" : np.int64,
                "esriFieldTypeGeometry" : object,
                "esriFieldTypeBlob" : object,
                "esriFieldTypeRaster" : object,
                "esriFieldTypeGUID" : str,
                "esriFieldTypeGlobalID" : str,
                "esriFieldTypeXML" : object
            }
        else:
            from datetime import datetime as _datetime
            _fld_lu = {
                "esriFieldTypeSmallInteger" : np.int32,
                "esriFieldTypeInteger" : np.int64,
                "esriFieldTypeSingle" : np.int32,
                "esriFieldTypeDouble" : float,
                "esriFieldTypeFloat" : float,
                "esriFieldTypeString" : str,
                "esriFieldTypeDate" : _datetime,
                "esriFieldTypeOID" : np.int64,
                "esriFieldTypeGeometry" : object,
                "esriFieldTypeBlob" : object,
                "esriFieldTypeRaster" : object,
                "esriFieldTypeGUID" : str,
                "esriFieldTypeGlobalID" : str,
                "esriFieldTypeXML" : object
            }
        def feature_to_row(feature, sr):
            """:return: a feature from a dict"""
            from arcgis.geometry import Geometry
            geom = feature['geometry'] if 'geometry' in feature else None
            attribs = feature['attributes'] if 'attributes' in feature else {}
            if 'centroid' in feature:
                if attribs is None:
                    attribs = {'centroid' : feature['centroid']}
                elif 'centroid' in attribs:
                    import uuid
                    fld = "centroid_" + uuid.uuid4().hex[:2]
                    attribs[fld] = feature['centroid']
                else:
                    attribs['centroid'] = feature['centroid']
            if geom:
                if "spatialReference" not in geom:
                    geom["spatialReference"] = sr
                attribs['SHAPE'] = Geometry(geom)
            return attribs
        #------------------------------------------------------------------
        try:
            featureset_dict = self._con.post(url, params,
                                             token=self._token)
        except Exception as queryException:
            error_list = ["Error performing query operation", "HTTP Error 504: GATEWAY_TIMEOUT"]
            if any(ele in queryException.__str__() for ele in error_list):
                # half the max record count
                max_record = int(params['resultRecordCount']) if 'resultRecordCount' in params else 1000
                offset = int(params['resultOffset']) if 'resultOffset' in params else 0
                # reduce this number to 125 if you still sees 500/504 error
                if max_record < 250:
                    # when max_record is lower than 250, but still getting error 500 or 504, just exit with exception
                    raise queryException
                else:
                    max_rec = int((max_record + 1) / 2)
                    i = 0
                    featureset_dict = None
                    while max_rec * i < max_record:
                        params['resultRecordCount'] = max_rec if max_rec * (i + 1) <= max_record else (
                                    max_record - max_rec * i)
                        params['resultOffset'] = offset + max_rec * i
                        try:
                            records = self._query(url, params, raw=True)
                            if featureset_dict is not None:
                                for feature in records['features']:
                                    featureset_dict['features'].append(feature)
                            else:
                                featureset_dict = records
                            i += 1
                        except Exception as queryException2:
                            raise queryException2

            else:
                raise queryException

        if len(featureset_dict['features']) == 0:
            return pd.DataFrame([])
        sr = None
        if 'spatialReference' in featureset_dict:
            sr = featureset_dict['spatialReference']

        df = None
        dtypes = None
        geom = None
        names = None
        dfields = []
        rows = [feature_to_row(row, sr) \
                for row in featureset_dict['features']]
        if len(rows) == 0:
            return None
        df = pd.DataFrame.from_records(data=rows)
        if 'fields' in featureset_dict:
            dtypes = {}
            names = []
            fields = featureset_dict['fields']
            for fld in fields:
                if fld['type'] != "esriFieldTypeGeometry":
                    dtypes[fld['name']] = _fld_lu[fld['type']]
                    names.append(fld['name'])
                if fld['type'] == 'esriFieldTypeDate':
                    dfields.append(fld['name'])
        if 'SHAPE' in featureset_dict:
            df.spatial.set_geometry('SHAPE')
        if len(dfields) > 0:
            df[dfields] = df[dfields].apply(pd.to_datetime, unit='ms')
        return df

###########################################################################
class MapRasterLayer(MapFeatureLayer):
    """
    A Map Raster Layer represents a geo-referenced image hosted in a Map Service.
    """
    @property
    def _lyr_dict(self):
        url = self.url

        if "lods" in self.container.properties:
            lyr_dict =  { 'type' : 'ArcGISTiledMapServiceLayer', 'url' : url }

        else:
            lyr_dict =  { 'type' : type(self.container).__name__, 'url' : url }

        if self._token is not None:
            lyr_dict['serviceToken'] = self._token

        if self.filter is not None:
            lyr_dict['filter'] = self.filter
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict

    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += '?token=' + self._token

        if "lods" in self.container.properties:
            lyr_dict =  { 'type' : 'ArcGISTiledMapServiceLayer', 'url' : self.container.url }

        else:
            lyr_dict =  { 'type' : type(self.container).__name__, 'url' : self.container.url }

        if self.filter is not None:
            lyr_dict['options'] = json.dumps({ "definition_expression": self.filter })
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
###########################################################################
class MapTable(MapFeatureLayer):
    """
    Tables represent entity classes with uniform properties. In addition to working with "entities with location" as
    features, the GIS can also work with non-spatial entities as rows in tables.

    Working with tables is similar to working with map feature layers, except that the rows (Features) in a table do not
    have a geometry, and tables ignore any geometry related operation.
    """

    @classmethod
    def fromitem(cls, item, table_id=0):
        """
        Creates a Table from a GIS Item.
        The type of item should be a 'Map Image Service'.
        The layer_id is the id of the layer.
        """
        return item.tables[table_id]

    #----------------------------------------------------------------------
    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict =  { 'type' : "FeatureLayer",
                      'url' : url }
        if self._token is not None:
            lyr_dict['serviceToken'] = self._token

        if self.filter is not None:
            lyr_dict['filter'] = self.filter
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += '?token=' + self._token

        lyr_dict = {'type': "FeatureLayer", 'url': url}

        if self.filter is not None:
            lyr_dict['options'] = json.dumps({ "definition_expression": self.filter })
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
    #----------------------------------------------------------------------
    def query(self,
              where="1=1",
              out_fields="*",
              time_filter=None,
              return_count_only=False,
              return_ids_only=False,
              return_distinct_values=False,
              group_by_fields_for_statistics=None,
              statistic_filter=None,
              result_offset=None,
              result_record_count=None,
              object_ids=None,
              gdb_version=None,
              order_by_fields=None,
              out_statistics=None,
              return_all_records=True,
              historic_moment=None,
              sql_format=None,
              return_exceeded_limit_features=None,
              as_df=False,
              range_values=None,
              parameter_values=None,
              **kwargs):
        """
        Queries a Table Layer based on a set of criteria.

        ===============================     ====================================================================
        **Argument**                        **Description**
        -------------------------------     --------------------------------------------------------------------
        where                               Optional string. The default is 1=1. The selection sql statement.
        -------------------------------     --------------------------------------------------------------------
        out_fields                          Optional List of field names to return. Field names can be specified
                                            either as a List of field names or as a comma separated string.
                                            The default is "*", which returns all the fields.
        -------------------------------     --------------------------------------------------------------------
        object_ids                          Optional string. The object IDs of this layer or table to be queried.
                                            The object ID values should be a comma-separated string.
        -------------------------------     --------------------------------------------------------------------
        time_filter                         Optional list. The format is of [<startTime>, <endTime>] using
                                            datetime.date, datetime.datetime or timestamp in milliseconds.
                                            Syntax: time_filter=[<startTime>, <endTime>] ; specified as
                                                    datetime.date, datetime.datetime or timestamp in
                                                    milliseconds
        -------------------------------     --------------------------------------------------------------------
        gdb_version                         Optional string. The geodatabase version to query. This parameter
                                            applies only if the isDataVersioned property of the layer is true.
                                            If this is not specified, the query will apply to the published
                                            map's version.
        -------------------------------     --------------------------------------------------------------------
        return_geometry                     Optional boolean. If true, geometry is returned with the query.
                                            Default is true.
        -------------------------------     --------------------------------------------------------------------
        return_distinct_values              Optional boolean.  If true, it returns distinct values based on the
                                            fields specified in out_fields. This parameter applies only if the
                                            supportsAdvancedQueries property of the layer is true.
        -------------------------------     --------------------------------------------------------------------
        return_ids_only                     Optional boolean. Default is False.  If true, the response only
                                            includes an array of object IDs. Otherwise, the response is a
                                            feature set.
        -------------------------------     --------------------------------------------------------------------
        return_count_only                   Optional boolean. If true, the response only includes the count
                                            (number of features/records) that would be returned by a query.
                                            Otherwise, the response is a feature set. The default is false. This
                                            option supersedes the returnIdsOnly parameter. If
                                            returnCountOnly = true, the response will return both the count and
                                            the extent.
        -------------------------------     --------------------------------------------------------------------
        order_by_fields                     Optional string. One or more field names on which the
                                            features/records need to be ordered. Use ASC or DESC for ascending
                                            or descending, respectively, following every field to control the
                                            ordering.
                                            example: STATE_NAME ASC, RACE DESC, GENDER
        -------------------------------     --------------------------------------------------------------------
        group_by_fields_for_statistics      Optional string. One or more field names on which the values need to
                                            be grouped for calculating the statistics.
                                            example: STATE_NAME, GENDER
        -------------------------------     --------------------------------------------------------------------
        out_statistics                      Optional string. The definitions for one or more field-based
                                            statistics to be calculated.

                                            Syntax:

                                            [
                                                {
                                                  "statisticType": "<count | sum | min | max | avg | stddev | var>",
                                                  "onStatisticField": "Field1",
                                                  "outStatisticFieldName": "Out_Field_Name1"
                                                },
                                                {
                                                  "statisticType": "<count | sum | min | max | avg | stddev | var>",
                                                  "onStatisticField": "Field2",
                                                  "outStatisticFieldName": "Out_Field_Name2"
                                                }
                                            ]
        -------------------------------     --------------------------------------------------------------------
        result_offset                       Optional integer. This option can be used for fetching query results
                                            by skipping the specified number of records and starting from the
                                            next record (that is, resultOffset + 1th). This option is ignored
                                            if return_all_records is True (i.e. by default).
        -------------------------------     --------------------------------------------------------------------
        result_record_count                 Optional integer. This option can be used for fetching query results
                                            up to the result_record_count specified. When result_offset is
                                            specified but this parameter is not, the map service defaults it to
                                            max_record_count. The maximum value for this parameter is the value
                                            of the layer's max_record_count property. This option is ignored if
                                            return_all_records is True (i.e. by default).
        -------------------------------     --------------------------------------------------------------------
        return_all_records                  Optional boolean. When True, the query operation will call the
                                            service until all records that satisfy the where_clause are
                                            returned. Note: result_offset and result_record_count will be
                                            ignored if return_all_records is True. Also, if return_count_only,
                                            return_ids_only, or return_extent_only are True, this parameter
                                            will be ignored.
        -------------------------------     --------------------------------------------------------------------
        historic_moment                     Optional integer. The historic moment to query. This parameter
                                            applies only if the layer is archiving enabled and the
                                            supportsQueryWithHistoricMoment property is set to true. This
                                            property is provided in the layer resource.

                                            If historic_moment is not specified, the query will apply to the
                                            current features.
        -------------------------------     --------------------------------------------------------------------
        sql_format                          Optional string.  The sql_format parameter can be either standard
                                            SQL92 standard or it can use the native SQL of the underlying
                                            datastore native. The default is none which means the sql_format
                                            depends on useStandardizedQuery parameter.
                                            Values: none | standard | native
        -------------------------------     --------------------------------------------------------------------
        return_exceeded_limit_features      Optional boolean. Optional parameter which is true by default. When
                                            set to true, features are returned even when the results include
                                            'exceededTransferLimit': True.

                                            When set to false and querying with resultType = tile features are
                                            not returned when the results include 'exceededTransferLimit': True.
                                            This allows a client to find the resolution in which the transfer
                                            limit is no longer exceeded without making multiple calls.
        -------------------------------     --------------------------------------------------------------------
        as_df                               Optional boolean.  If True, the results are returned as a DataFrame
                                            instead of a FeatureSet.
        -------------------------------     --------------------------------------------------------------------
        range_values                        Optional List. Allows you to filter features from the layer that are
                                            within the specified range instant or extent.

                                            :Syntax:

                                            [
                                                {
                                                  "name": "range name",
                                                  "value": <value> or [ <value1>, <value2> ]  # single value or a value-range
                                                      # None is allowed in value-range case -- that means infinity
                                                      # e.g. [None, 1500] means all features with values <= 1500
                                                      # [1000, None] means all features with values >= 1000
                                                  },
                                                  {
                                                    "name": "range name 2",
                                                    "value": <value> or  [ <value3>, <value4> ]
                                                  }
                                                }
                                            ]

        -------------------------------     --------------------------------------------------------------------
        parameter_values                    Optional Dict. Allows you to filter the features layers by specifying
                                            value(s) to an array of pre-authored parameterized filters for those
                                            layers. When value is not specified for any parameter in a request,
                                            the default value, that is assigned during authoring time, gets used
                                            instead.

                                            When a parameterInfo allows multiple values, you must pass them in an array.

                                            Note: Check parameterInfos at the layer resources for the available
                                            parameterized filters, their default values and expected data type.
        -------------------------------     --------------------------------------------------------------------
        kwargs                              Optional dict. Optional parameters that can be passed to the Query
                                            function.  This will allow users to pass additional parameters not
                                            explicitly implemented on the function. A complete list of functions
                                            available is documented on the Query REST API.
        ===============================     ====================================================================

        :returns: A FeatureSet or Panda's DataFrame containing the features matching the query unless another return type is specified, such as count
        """
        as_raw = as_df
        if self._dynamic_layer is None:
            url = self._url + "/query"
        else:
            url = "%s/query" % self._url.split('?')[0]

        params = {"f": "json"}
        if self._dynamic_layer is not None:
            params['layer'] = self._dynamic_layer
        if historic_moment is not None:
            params['historicMoment'] = historic_moment
        if sql_format is not None:
            params['sqlFormat'] = sql_format
        if return_exceeded_limit_features is not None:
            params['returnExceededLimitFeatures'] = return_exceeded_limit_features
        params['where'] = where
        params['returnDistinctValues'] = return_distinct_values
        params['returnCountOnly'] = return_count_only
        params['returnIdsOnly'] = return_ids_only

        # convert out_fields to a comma separated string
        if isinstance(out_fields, (list, tuple)):
            out_fields = ','.join(out_fields)

        if out_fields != '*' and not return_distinct_values:
            try:
                # Check if object id field is in out_fields.
                # If it isn't, add it
                object_id_field = [x.name for x in self.properties.fields if x.type == "esriFieldTypeOID"][0]
                if object_id_field not in out_fields.split(','):
                    out_fields = object_id_field + "," + out_fields
            except (IndexError, AttributeError):
                pass
        params['outFields'] = out_fields
        if return_count_only or return_ids_only:
            return_all_records = False
        if result_record_count and not return_all_records:
            params['resultRecordCount'] = result_record_count
        if result_offset and not return_all_records:
            params['resultOffset'] = result_offset
        if order_by_fields:
            params['orderByFields'] = order_by_fields
        if group_by_fields_for_statistics:
            params['groupByFieldsForStatistics'] = group_by_fields_for_statistics
        if statistic_filter and \
                isinstance(statistic_filter, StatisticFilter):
            params['outStatistics'] = statistic_filter.filter
        if out_statistics:
            params['outStatistics'] = out_statistics
        if gdb_version:
            params['gdbVersion'] = gdb_version
        if object_ids:
            params['objectIds'] = object_ids

        if time_filter is None and self.time_filter:
            params['time'] = self.time_filter
        elif time_filter is not None:
            if type(time_filter) is list:
                starttime = _date_handler(time_filter[0])
                endtime = _date_handler(time_filter[1])
                if starttime is None:
                    starttime = 'null'
                if endtime is None:
                    endtime = 'null'
                params['time'] = "%s,%s" % (starttime, endtime)
            elif isinstance(time_filter, dict):
                for key, val in time_filter.items():
                    params[key] = val
            else:
                params['time'] = _date_handler(time_filter)

        if len(kwargs) > 0:
            for key, val in kwargs.items():
                if key in ('returnCountOnly', 'returnIdsOnly') and val:
                    # If these keys are passed in as kwargs instead of parameters, set return_all_records
                    return_all_records = False
                params[key] = val
                del key, val

        if not return_all_records or "outStatistics" in params:
            if as_df:
                return self._query_df(url, params)
            return self._query(url, params, raw=as_raw)

        params['returnCountOnly'] = True
        record_count = self._query(url, params, raw=as_raw)
        if 'maxRecordCount' in self.properties:
            max_records = self.properties['maxRecordCount']
        else:
            max_records = 1000

        supports_pagination = True
        if ('advancedQueryCapabilities' not in self.properties or \
                'supportsPagination' not in self.properties['advancedQueryCapabilities'] or \
                not self.properties['advancedQueryCapabilities']['supportsPagination']):
            supports_pagination = False

        params['returnCountOnly'] = False
        if record_count == 0 and as_df:
            from arcgis.features.geo._array import GeoArray
            import numpy as np
            import pandas as pd
            _fld_lu = {
                "esriFieldTypeSmallInteger" : np.int32,
                "esriFieldTypeInteger" : np.int64,
                "esriFieldTypeSingle" : np.int32,
                "esriFieldTypeDouble" : float,
                "esriFieldTypeFloat" : float,
                "esriFieldTypeString" : str,
                "esriFieldTypeDate" : np.datetime64,
                "esriFieldTypeOID" : np.int64,
                "esriFieldTypeGeometry" : object,
                "esriFieldTypeBlob" : object,
                "esriFieldTypeRaster" : object,
                "esriFieldTypeGUID" : str,
                "esriFieldTypeGlobalID" : str,
                "esriFieldTypeXML" : object
            }
            columns = {}
            for fld in self.properties.fields:
                fld = dict(fld)
                columns[fld['name']] = _fld_lu[fld['type']]
            if "geometryType" in self.properties and \
               not self.properties.geometryType is None:
                columns['SHAPE'] = object
            df = pd.DataFrame([], columns=columns.keys()).astype(columns, True)
            if 'SHAPE' in df.columns:
                df['SHAPE'] = GeoArray([])
                df.spatial.set_geometry("SHAPE")
                df.spatial.renderer = self.renderer
                df.spatial._meta.source = self
            return df
        elif record_count <= max_records:
            if supports_pagination and record_count > 0:
                params['resultRecordCount'] = record_count
            if as_df:
                import pandas as pd
                df = self._query_df(url, params)
                dt_fields = [fld['name'] for fld in self.properties.fields \
                             if fld['type'] == 'esriFieldTypeDate']
                if 'SHAPE' in df.columns:
                    df.spatial.set_geometry('SHAPE')
                    df.spatial.renderer = self.renderer
                    df.spatial._meta.source = self
                for fld in dt_fields:
                    try:
                        if fld in df.columns:
                            df[fld] = pd.to_datetime(df[fld]/1000,
                                                 infer_datetime_format=True,
                                                 unit='s')
                    except:
                        if fld in df.columns:
                            df[fld] = pd.to_datetime(df[fld], infer_datetime_format=True)
                return df

            return self._query(url, params, raw=as_raw)

        result = None
        i = 0
        count = 0
        df = None
        dfs = []
        if not supports_pagination:
            params['returnIdsOnly'] = True
            oid_info = self._query(url, params, raw=as_raw)
            params['returnIdsOnly'] = False
            for ids in chunks(oid_info['objectIds'], max_records):
                ids = [str(i) for i in ids]
                sql = "%s in (%s)" % (oid_info['objectIdFieldName'], ",".join(ids))
                params['where'] = sql
                if not as_df:
                    records = self._query(url, params, raw=as_raw)
                    if result:
                        if 'features' in result:
                            result['features'].append(records['features'])
                        else:
                            result.features.extend(records.features)
                    else:
                        result = records
                else:
                    df = self._query_df(url, params)
                    dfs.append(df)
        else:
            while True:
                params['resultRecordCount'] = max_records
                params['resultOffset'] = max_records * i
                if not as_df:
                    records = self._query(url, params, raw=as_raw)

                    if result:
                        if 'features' in result:
                            result['features'].append(records['features'])
                        else:
                            result.features.extend(records.features)
                    else:
                        result = records

                    if len(records.features) < max_records:
                        break
                else:

                    df = self._query_df(url, params)
                    count += len(df)
                    dfs.append(df)
                    if count == record_count:
                        break
                i += 1
        if as_df:
            import pandas as pd
            dt_fields = [fld['name'] for fld in self.properties.fields \
                         if fld['type'] == 'esriFieldTypeDate']
            if len(dfs) == 1:
                df = dfs[0]
            else:
                df = pd.concat(dfs, sort=True)
                df.reset_index(drop=True, inplace=True)
            if 'SHAPE' in df.columns:
                df.spatial.set_geometry('SHAPE')
                df.spatial.renderer = self.renderer
                df.spatial._meta.source = self
            for fld in dt_fields:
                try:
                    df[fld] = pd.to_datetime(df[fld]/1000,
                                             infer_datetime_format=True,
                                             unit='s')
                except:
                    df[fld] = pd.to_datetime(df[fld], infer_datetime_format=True)
            return df
        return result



###########################################################################
class _MSILayerFactory(type):
    """
    Factory that generates the Map Service Layers

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /MapServer/<index>
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a Map Service Layer object

        from arcgis.mapping import SceneLayer
        ms_layer = MapServiceLayer(url='https://your_portal.com/arcgis/rest/services/service_name/MapServer/0')

        type(ms_layer)
        >> arcgis.mapping._types.MapTable

        print(s_layer.properties.name)
        >> 'pipe_properties'
    """
    def __call__(cls,
                 url,
                 gis=None,
                 container=None,
                 dynamic_layer=None):
        lyr = Layer(url=url, gis=gis)
        props = lyr.properties
        if 'type' in props and \
           props.type.lower() == 'table':
            return MapTable(url=url, gis=gis, container=container, dynamic_layer=container)
        elif 'type' in props and \
           props.type.lower() == 'raster layer':
            return MapRasterLayer(url=url, gis=gis, container=container, dynamic_layer=container)
        elif 'type' in props and \
           props.type.lower() == 'feature layer':
            return MapFeatureLayer(url=url, gis=gis, container=container, dynamic_layer=container)
        return lyr
###########################################################################
class MapServiceLayer(Layer, metaclass=_MSILayerFactory):
    """
    Factory that generates the Map Service Layers

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /MapServer/<index>
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS` object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a Map Service Layer object

        from arcgis.mapping import MapServiceLayer
        ms_layer = MapServiceLayer(url='https://your_portal.com/arcgis/rest/services/service_name/MapServer/0')

        type(ms_layer)
        >> arcgis.mapping._types.MapTable

        print(ms_layer.properties.name)
        >> 'pipe_properties'

    """
    def __init__(self, url, gis=None, container=None, dynamic_layer=None):
        """
        Constructs a Map Services Layer given a URL and GIS
        """
        super(MapServiceLayer, self).__init__(url=url, gis=gis,
                                              container=container,
                                              dynamic_layer=container)




import json
from pathlib import Path
from ._codetemplate import image_classifier_prf
from ._arcgis_model import _EmptyData
from functools import partial
import math
import types
from .._data import _raise_fastai_import_error  
import traceback
import logging
logger = logging.getLogger()

try:
    from ._arcgis_model import ArcGISModel, SaveModelCallback, _set_multigpu_callback, _resnet_family, _set_ddp_multigpu, _isnotebook
    import torch
    from torchvision import models
    from fastai.vision.learner import unet_learner, cnn_config
    import numpy as np
    from fastai.layers import CrossEntropyFlat
    from .._utils.segmentation_loss_functions import  FocalLoss, MixUpCallback, DiceLoss
    from ._unet_utils import is_no_color, LabelCallback, _class_array_to_rbg, predict_batch, show_results_multispectral
    from fastai.callbacks import EarlyStoppingCallback
    from torch.nn import Module as NnModule
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from .._utils.classified_tiles import per_class_metrics
    from ._psp_utils import accuracy
    from ._deeplab_utils import compute_miou
    import os as arcgis_os
    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    class NnModule():
        pass
    HAS_FASTAI = False


class UnetClassifier(ArcGISModel):
    """
    Creates a Unet like classifier based on given pretrained encoder.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `UnetClassifier`, which
                            is `resnet34` by default.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    backend                 Optional string. Controls the backend framework to be used
                            for this model, which is 'pytorch' by default.

                            valid options are 'pytorch', 'tensorflow'
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    class_balancing         Optional boolean. If True, it will balance the
                            cross-entropy loss inverse to the frequency
                            of pixels per class. Default: False. 
    ---------------------   -------------------------------------------
    mixup                   Optional boolean. If True, it will use mixup
                            augmentation and mixup loss. Default: False
    ---------------------   -------------------------------------------
    focal_loss              Optional boolean. If True, it will use focal loss
                            Default: False
    ---------------------   -------------------------------------------
    dice_loss_fraction      Optional float. 
                            Min_val=0, Max_val=1 
                            If > 0 , model will use a combination of default or 
                            focal(if focal=True) loss with the specified fraction 
                            of dice loss.
                            E.g. 
                            for dice = 0.3, loss = (1-0.3)*default loss + 0.3*dice
                            Default: 0
    ---------------------   -------------------------------------------
    ignore_classes          Optional list. It will contain the list of class
                            values on which model will not incur loss.
                            Default: []
    =====================   ===========================================

    :returns: `UnetClassifier` Object
    """

    def __init__(self, data, backbone=None, pretrained_path=None, backend='pytorch', *args, **kwargs):

        self._backend = backend
        if self._backend == 'tensorflow':
            super().__init__(data, None)
            self._intialize_tensorflow(data, backbone, pretrained_path, kwargs)
        else:
            super().__init__(data, backbone, **kwargs)

            self._check_dataset_support(self._data)
            if not (self._check_backbone_support(getattr(self, '_backbone', backbone))):
                raise Exception(f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")

            self._ignore_classes = kwargs.get('ignore_classes', [])
            if self._ignore_classes != [] and len(data.classes) <= 3:
                raise Exception(f"`ignore_classes` parameter can only be used when the dataset has more than 2 classes.")

            data_classes = list(self._data.class_mapping.keys())
            if 0 not in list(data.class_mapping.values()):
                self._ignore_mapped_class = [data_classes.index(k) + 1 for k in self._ignore_classes if k != 0]
            else:
                self._ignore_mapped_class = [data_classes.index(k) + 1 for k in self._ignore_classes]
            if self._ignore_classes != []:
                if 0 not in self._ignore_mapped_class:
                    self._ignore_mapped_class.insert(0, 0)
                global accuracy
                accuracy = partial(accuracy, ignore_mapped_class=self._ignore_mapped_class)       

            self.mixup = kwargs.get('mixup', False)
            self.class_balancing = kwargs.get('class_balancing', False)
            self.focal_loss = kwargs.get('focal_loss', False)
            self.dice_loss_fraction = kwargs.get('dice_loss_fraction', False)
            self.weighted_dice = kwargs.get('weighted_dice', False)

            self._code = image_classifier_prf

            backbone_cut = None
            backbone_split = None

            if hasattr(self, '_orig_backbone'):
                _backbone_meta = cnn_config(self._orig_backbone)
                backbone_cut = _backbone_meta['cut']
                backbone_split = _backbone_meta['split']

            if not _isnotebook() and arcgis_os.name=='posix':
                _set_ddp_multigpu(self)
                if self._multigpu_training:
                    self.learn = unet_learner(data, arch=self._backbone, metrics=accuracy, wd=1e-2, bottle=True, last_cross=True, cut=backbone_cut, split_on=backbone_split).to_distributed(self._rank_distributed)
                else:
                    self.learn = unet_learner(data, arch=self._backbone, metrics=accuracy, wd=1e-2, bottle=True, last_cross=True, cut=backbone_cut, split_on=backbone_split)
            else:
                self.learn = unet_learner(data, arch=self._backbone, metrics=accuracy, wd=1e-2, bottle=True, last_cross=True, cut=backbone_cut, split_on=backbone_split)

            class_weight = None
            if self.class_balancing:
                if data.class_weight is not None:
                    # Handle condition when nodata is already at pixel value 0 in data
                    if (data.c-1) == data.class_weight.shape[0]:
                        class_weight = torch.tensor([data.class_weight.mean()] + data.class_weight.tolist()).float().to(self._device)
                    else:
                        class_weight = torch.tensor(data.class_weight).float().to(self._device)
                else:
                    if getattr(data, 'overflow_encountered', False):
                        logger.warning("Overflow Encountered. Ignoring `class_balancing` parameter.")
                        class_weight = [1] * len(data.classes)
                    else:
                        logger.warning("Could not find 'NumPixelsPerClass' in 'esri_accumulated_stats.json'. Ignoring `class_balancing` parameter.")


            if self._ignore_classes != []:
                if not self.class_balancing:
                    class_weight = torch.tensor([1] * data.c).float().to(self._device)
                class_weight[self._ignore_mapped_class] = 0.

            self._final_class_weight = class_weight
            self.learn.loss_func = CrossEntropyFlat(class_weight, axis=1)

            if self.focal_loss:
                self.learn.loss_func = FocalLoss(self.learn.loss_func)
            if self.mixup:
                self.learn.callbacks.append(MixUpCallback(self.learn))

            if self.dice_loss_fraction:
                self.learn.loss_func = DiceLoss(self.learn.loss_func, self.dice_loss_fraction,  weighted_dice=self.weighted_dice)

            self._arcgis_init_callback() # make first conv weights learnable
            self.learn.callbacks.append(LabelCallback(self.learn))  #appending label callback

            self.learn.model = self.learn.model.to(self._device)
            # _set_multigpu_callback(self) # MultiGPU doesn't work for U-Net. (Fastai-Forums)
            if pretrained_path is not None:
                self.load(pretrained_path)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @property
    def supported_backbones(self):
        """ Supported torchvision backbones for this model. """        
        return UnetClassifier._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family]

    @property
    def  supported_datasets(self):
        """ Supported dataset types for this model. """
        return UnetClassifier._supported_datasets()
    
    @staticmethod
    def _supported_datasets():
        return ['Classified_Tiles']    

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a Unet like classifier from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================
        
        :returns: `UnetClassifier` Object
        """
        return cls.from_emd(data, emd_path)

    @classmethod
    def from_emd(cls, data, emd_path):
        """
        Creates a Unet like classifier from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        =====================   ===========================================
        
        :returns: `UnetClassifier` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)
            
        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd['ModelFile'])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd['ModelParameters']

        try:
            class_mapping = {i['Value']: i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value']: i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue']: i['ClassName'] for i in emd['Classes']}
            color_mapping = {i['ClassValue']: i['Color'] for i in emd['Classes']}

        resize_to = emd.get('resize_to')

        if data is None:
            data = _EmptyData(path=emd_path.parent.parent, loss_func=None, c=len(class_mapping) + 1,
                              chip_size=emd['ImageHeight'])
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data = get_multispectral_data_params_from_emd(data, emd)
                
            data.emd_path = emd_path
            data.emd = emd
            data._is_empty = True

        data.resize_to = resize_to        

        return cls(data, **model_params, pretrained_path=str(model_file))

    @property
    def _model_metrics(self):
        return {'accuracy': '{0:1.4e}'.format(self._get_model_metrics())}
        
    def _get_emd_params(self, save_inference_file):
        import random
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_unet"
        _emd_template["ModelType"] = "ImageClassification"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
        else:
            _emd_template["InferenceFunction"] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageClassifier.py"
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template["ignore_mapped_class"] = self._ignore_mapped_class
        _emd_template["SupportsVariableTileSize"] = True

        _emd_template['Classes'] = []
        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)] if is_no_color(self._data.color_mapping) else \
                self._data.color_mapping[inverse_class_mapping[class_name]]
            class_data["Color"] = color
            _emd_template['Classes'].append(class_data.copy())

        return _emd_template

    def _predict_batch(self, imagetensor_batch):
        return predict_batch(self, imagetensor_batch)

    def _show_results_multispectral(self, rows=5, alpha=0.7, **kwargs): # parameters adjusted in kwargs
        return_fig = kwargs.get('return_fig', False)
        ret_val = show_results_multispectral(
            self, 
            nrows=rows, 
            alpha=alpha, 
            **kwargs
        )
        if return_fig:
            fig, ax = ret_val
            return fig

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        self._check_requisites()
        self.learn.callbacks = [x for x in self.learn.callbacks if not isinstance(x, LabelCallback)]
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        self.learn.show_results(rows=rows, ignore_mapped_class=self._ignore_mapped_class, **kwargs)

    def accuracy(self):
        try:
            return self.learn.validate()[1].tolist()
        except Exception as e:
            accuracy = self._data.emd.get('accuracy')
            if accuracy:
                return accuracy
            else:
                logger.error("Metric not found in the loaded model")

    def _get_model_metrics(self, **kwargs):
        checkpoint = kwargs.get('checkpoint', True)
        if not hasattr(self.learn, 'recorder'):
            return 0.0

        try:
            model_accuracy = self.learn.recorder.metrics[-1][0]
            if checkpoint:
                val_losses = self.learn.recorder.val_losses
                model_accuracy = self.learn.recorder.metrics[val_losses.index(min(val_losses))][0]
        except:
            model_accuracy = 0.0

        return float(model_accuracy)

    def mIOU(self, mean=False, show_progress=True):

        """
        Computes mean IOU on the validation set for each class.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                mean IOU, otherwise returns mean iou of all
                                classes combined.   
        ---------------------   -------------------------------------------
        show_progress           Optional bool. Displays the progress bar if
                                True.                     
        =====================   ===========================================
        
        :returns: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()
        num_classes = torch.arange(self._data.c)
        miou = compute_miou(self, self._data.valid_dl, mean, num_classes, show_progress, self._ignore_mapped_class)
        if mean:
            miou = [miou[i] for i in range(len(miou)) if i not in self._ignore_mapped_class]
            return np.mean(miou)
        if self._ignore_mapped_class == []:
            return dict(zip(['0'] + self._data.classes[1:], miou))
        else:
            class_values = [0] + list(self._data.class_mapping.keys())
            return {class_values[i]: miou[i] for i in range(len(miou)) if i not in self._ignore_mapped_class} 

    ## Tensorflow specific functions start ##
    def _intialize_tensorflow(self, data, backbone, pretrained_path, kwargs):
        self._check_tf()
        self._ignore_mapped_class = []
        
        import tensorflow as tf
        from .._utils.common import get_color_array
        from .._utils.common_tf import handle_backbone_parameter, get_input_shape
        from .._model_archs.unet_tf import get_unet_tf_model 
        from tensorflow.keras.losses import SparseCategoricalCrossentropy, BinaryCrossentropy
        from .._utils.fastai_tf_fit import TfLearner, defaults
        from tensorflow.keras.models import Model
        from tensorflow.keras.optimizers import Adam
        from .._utils.common import kwarg_fill_none
        
        if data._is_multispectral:
            raise Exception('Multispectral data is not supported with backend="tensorflow"')

        # Intialize Tensorflow
        self._init_tensorflow(data, backbone)

        # Loss Function
        #self._loss_function_tf_ = BinaryCrossentropy(from_logits=True)
        self._loss_function_tf_ = SparseCategoricalCrossentropy(from_logits=True, reduction='auto')

        self._mobile_optimized = kwarg_fill_none(kwargs, 'mobile_optimized', self._backbone_mobile_optimized)

        # Create Unet Model
        model = get_unet_tf_model(
            self._backbone_initalized,
            data,
            mobile_optimized=self._mobile_optimized
        )

        self.learn = TfLearner(
            data, 
            model,
            opt_func=Adam,
            loss_func=self._loss_function_tf,
            true_wd=True, 
            bn_wd=True, 
            wd=defaults.wd, 
            train_bn=True
        )
        
        self.learn.unfreeze()
        self.learn.freeze_to(len(self._backbone_initalized.layers))

        self.show_results = self._show_results_multispectral

        self._code = image_classifier_prf

    def _loss_function_tf(self, target, predictions):
        import tensorflow as tf
        # print(target.shape, predictions.shape)
        # print(target.dtype, predictions.dtype)
        # print(tf.unique(tf.reshape(target, [-1]))[0])
        # print('\n', tf.unique(tf.reshape(predictions, [-1]))[0])
        #print(tf.unique(tf.reshape(target, [-1])).numpy(), tf.unique(tf.reshape(predictions, [-1])))
        target = tf.squeeze(target, axis=1)

        # from .._utils.pixel_classification import segmentation_mask_to_one_hot
        # from .._utils.fastai_tf_fit import _pytorch_to_tf
        # target = _pytorch_to_tf(segmentation_mask_to_one_hot(target.cpu().numpy(), self._data.c).permute(0, 2, 3, 1))

        return self._loss_function_tf_(target, predictions)

    ## Tensorflow specific functions end ##

    def per_class_metrics(self, ignore_classes=[]):
        """
        Computer per class precision, recall and f1-score on validation set.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        self                    segmentation model object -> [PSPNetClassifier | UnetClassifier | DeepLab]
        ---------------------   -------------------------------------------
        ignore_classes          Optional list. It will contain the list of class
                                values on which model will not incur loss.
                                Default: []    
        =====================   ===========================================

        Returns per class precision, recall and f1 scores 
        """
        try:
            self._check_requisites()
            ## Calling imported function `per_class_metrics`
            return per_class_metrics(self, ignore_classes)
        except:
            import pandas as pd
            return pd.read_json(self._data.emd['per_class_metrics'])


import os
import json
import traceback

HAS_TRANSFORMER = True

try:
    import torch
    from transformers import AutoModelForSequenceClassification
    from ._arcgis_transformer import ArcGISTransformer, transformer_seq_length
except Exception as e:
    transformer_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_TRANSFORMER = False
    transformer_seq_length = 512
    class ArcGISTransformer:
        pass

backbone_models_map = {
    'bert': ('bert-base-cased', 'bert-base-uncased', 'bert-large-cased', 'bert-large-uncased'),
    'albert': ('albert-base-v1', 'albert-base-v2', 'albert-large-v1', 'albert-large-v2',
               'albert-xlarge-v1', 'albert-xlarge-v2', 'albert-xxlarge-v1', 'albert-xxlarge-v2'),
    'roberta': ('roberta-base', 'roberta-large', 'distilroberta-base'),
    'distilbert': ('distilbert-base-cased', 'distilbert-base-uncased'),
    'xlnet': ('xlnet-base-cased', 'xlnet-large-cased'),
    'xlm': ('xlm-mlm-ende-1024', 'xlm-mlm-enfr-1024', 'xlm-mlm-xnli15-1024', 'xlm-mlm-en-2048'),
    'flaubert': ('flaubert/flaubert_small_cased', 'flaubert/flaubert_base_cased', 'flaubert/flaubert_base_uncased',
                 'flaubert/flaubert_large_cased'),
    'xlm-roberta': ('xlm-roberta-base', 'xlm-roberta-large'),
    'longformer': ('allenai/longformer-base-4096',),
    'mobilebert': ('google/mobilebert-uncased',),
    'electra': ('google/electra-base-discriminator', 'google/electra-base-generator'),
    'bart': ('facebook/bart-base', 'facebook/bart-large'),
    'camembert': ('camembert-base',)
}

transformer_architectures = ['BERT', 'RoBERTa', 'DistilBERT', 'ALBERT', 'FlauBERT', 'CamemBERT',
                             'XLNet', 'XLM', 'XLM-RoBERTa', 'Bart', 'ELECTRA', 'Longformer', 'MobileBERT']

backbone_models_reverse_map = {x:key for key, val in backbone_models_map.items() for x in val}


class TransformerForTextClassification(ArcGISTransformer):

    _supported_backbones = transformer_architectures

    def __init__(self, architecture, pretrained_model_name, config=None, pretrained_model_path=None,
                 seq_len=transformer_seq_length):
        super().__init__(architecture, pretrained_model_name, config, pretrained_model_path, task="classification")
        self._seq_len = seq_len
        self._max_seq_len = None

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % type(self).__name__

    def _process_config(self):
        """
        Process the model config to convert the class values to integers in the
        id2label attribute (which is a dictionary) of the model config
        """
        self._config.id2label = {int(x): y for x, y in self._config.id2label.items()}

    def _set_max_seq_length(self):
        """
        Set the max-seq-len parameter to avoid out of memory issues for transformers
        like Bart, Longformer, XLNet
        """
        self._max_seq_len = min(self._tokenizer.model_max_length, self._seq_len)

    @classmethod
    def _available_backbone_models(cls, architecture):
        """
        Provides a list of available models for a given transformer architecture

        =====================   =================================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------------
        architecture            Required string. The transformer architecture for
                                which we wish to get the available models
        ---------------------   -------------------------------------------------
        return: tuple containing available models for the given architecture
        """
        if architecture.lower() in backbone_models_map:
            return backbone_models_map[architecture.lower()]
        else:
            return f"Error, wrong architecture name - {architecture} supplied. " \
                   f"Please choose from {cls._supported_backbones}"

    def save(self, model_path):
        """
        Method to save the fine-tuned model to the disk

        =====================   =================================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------------
        model_path              Required string. The disk location where the
                                fine-tuned model has to be saved
        ---------------------   -------------------------------------------------
        """
        if not os.path.exists(model_path):
            os.mkdir(model_path)
        model_params = {"architecture": self._transformer_architecture,
                        "pretrained_model": self._transformer_pretrained_model_name}
        file_path = os.path.join(model_path, self._outfile)
        out_file = open(file_path, "w")
        json.dump(model_params, out_file)
        self._transformer.save_pretrained(model_path)
        # self._tokenizer.save_pretrained(model_path)

    @classmethod
    def load_pretrained_model(cls, path):
        """
        Method to load the fine-tuned model which was saved on the disk

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. The disk location from
                                where the fine-tuned model has to be loaded
        ---------------------   -------------------------------------------
        """
        architecture_file_path = os.path.join(path, cls._outfile)
        if os.path.exists(architecture_file_path):
            with open(architecture_file_path) as f:
                data = json.load(f)
            architecture = data.get("architecture")
            pretrained_model_name = data.get("pretrained_model")
            obj = TransformerForTextClassification(architecture, pretrained_model_name, pretrained_model_path=path)
            obj.init_model()
            return obj
        else:
            raise Exception(f"{cls._outfile} not present at {path}")

    def forward(self, input_ids, **kwargs):
        """
        Return only the logits from the transfomer model

        =====================   ==============================================
        **Argument**            **Description**
        ---------------------   ----------------------------------------------
        input_ids               tensor object. tensor containing the token-ids
                                got by calling tokenizer.encode method to the
                                input text. works same for mini-batches
        ---------------------   ----------------------------------------------

        :return: the logits from the transformer model
        """
        logits = self._transformer(input_ids)[0]
        return logits

    def _load_transformer(self):
        """
        Method to load the transformer model. For this class we want to load the
        transformer model for performing Sequence Classification tasks
        """
        if self._pretrained_model_path and os.path.exists(self._pretrained_model_path):
            base_path, extension = os.path.splitext(self._pretrained_model_path)
            path = base_path + ".pth"
            if extension == ".emd" and os.path.exists(path):
                self._transformer = AutoModelForSequenceClassification.from_pretrained(path, config=self._config)
            else:
                self._transformer = AutoModelForSequenceClassification.from_pretrained(self._pretrained_model_path)
        else:
            self._transformer = AutoModelForSequenceClassification.\
                from_pretrained(self._transformer_pretrained_model_name, config=self._config)

    def predict_class(self, text, device, is_multilabel_problem=False, thresh=None):
        """
        Method to predict the class labels for an input text

        =====================   ==============================================
        **Argument**            **Description**
        ---------------------   ----------------------------------------------
        text                    Required string. The text for which we wish to
                                predict the class
        ---------------------   ----------------------------------------------

        :return: the predicted label and the corresponding confidence score.
        """
        # device_type = next(self._transformer.parameters()).device.type
        # device = torch.device("cuda:0" if device_type == "cuda" else "cpu")
        sequence = torch.tensor([self._tokenizer.encode(text, max_length=self._max_seq_len, truncation=True)]).to(device)
        logits = self._transformer(sequence)[0]
        if is_multilabel_problem:
            results = torch.sigmoid(logits)[0]
            category, raw_pred = [], []
            for idx, x in enumerate(results):
                pred = 0
                if x.item() > thresh:
                    pred = 1
                    category.append(self._config.id2label[idx])
                raw_pred.append(pred)
            return ";".join(category), raw_pred, [round(x, 4) for x in results.tolist()]
        else:
            results = torch.softmax(logits, dim=1)[0]
            clas = self._config.id2label[torch.argmax(logits, dim=1)[0].item()]
            score = results[torch.argmax(logits, dim=1)[0]]
            return clas, score.item()



def extract_zipfile(filepath, filename, remove=False):
    """ Function to extract the contents of a zip file
        Args:
            filepath: absolute path to the file directory.
            filename: name of the zip file to be extracted.
            remove: default=False, removes the original zip file 
                    after extracting the contents if True
    """
    import os, zipfile
    with zipfile.ZipFile(os.path.join(filepath, filename),"r") as zip_ref:
        zip_ref.extractall(filepath)
    if remove:
        os.remove(os.path.join(filepath, filename))



try:
    from traitlets import Dict
except ImportError:
    class Dict:
        pass

class observablelist(list):
    """Acts just like a `list` primitive, but calls `on_change()`
    every time the list mutates."""
    def __setitem__(self, *args, **kwargs):
        old = list(self)
        super().__setitem__(*args, **kwargs)
        new = list(self)
        
        self._on_change(old, new)
        
    def __delitem__(self, *args, **kwargs):
        old = list(self)
        super().__delitem__(*args, **kwargs)
        new = list(self)
        
        self._on_change(old, new)

    def append(self, *args, **kwargs):
        old = list(self)
        super().append(*args, **kwargs)
        new = list(self)
        
        self._on_change(old, new)
 
    def extend(self, *args, **kwargs):
        old = list(self)
        super().extend(*args, **kwargs)
        new = list(self)
        
        self._on_change(old, new)

    def insert(self, *args, **kwargs):
        old = list(self)
        super().insert(*args, **kwargs)
        new = list(self)
        
        self._on_change(old, new)

    def remove(self, *args, **kwargs):
        old = list(self)
        super().remove(*args, **kwargs)
        new = list(self)
        
        self._on_change(old, new)
  
    def pop(self, *args, **kwargs):
        old = list(self)
        super().pop(*args, **kwargs)
        new = list(self)
  
        self._on_change(old, new)
    
    def clear(self, *args, **kwargs):
        old = list(self)
        super().clear(*args, **kwargs)
        new = list(self)
        
        self._on_change(old, new)

    def _on_change(self, old, new):
        """Called everytime the list is mutated (i.e. my_list.append('foo'))
        Overwrite this function with what you want your callback to be
        
        Args
        ----
        old: a copy of the list object before the mutation
        new: a copy of the list object after the mutation
        """
        pass

class observabledict(dict):
    """Acts just like a `dict` primitive, but calls `on_change()`
    every time the dict mutates."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._replace_children_with_observables(self)

    def _replace_children_with_observables(self, dict_):
        """For each nested value in a multi-layer dict object,
        replace all `dict` subchildren with `observabledict`,
        replace all `list` subchildren with `observablelist`
        """
        if isinstance(dict_, dict):
            for key in dict_:
                value = dict_[key]
                if isinstance(value, dict):
                    dict_[key] = observabledict(value)
                    dict_[key]._on_change = self._child_on_change
                if isinstance(value, list):
                    dict_[key] = observablelist(value)
                    dict_[key]._on_change = self._child_on_change

    def __setitem__(self, key, value):
        if isinstance(value, dict):
            value = observabledict(value)
            value._on_change = self._child_on_change
        elif isinstance(value, list):
            value = observablelist(value)
            value._on_change = self._child_on_change
        
        old = dict(self)
        super().__setitem__(key, value)
        new = dict(self)

        self._on_change(old, new)

    def __delitem__(self, *args, **kwargs):
        old = dict(self)
        super().__delitem__(*args, **kwargs)
        new = dict(self)
        
        self._on_change(old, new)

    def update(self, *args, **kwargs):
        old = dict(self)
        super().update(*args, **kwargs)
        new = dict(self)
        
        self._on_change(old, new)

    def _child_on_change(self, old_child, new_child):
        """Make any child `on_change` call bubble up the parent `on_change`,
        but with the correct old/new values. `self` has already been updated,
        so find what has changed and replace it with the old value
        """
        new = dict(self)
        old = dict(self)
        self._recurs_replace_child_with(old, new_child, old_child)

        self._on_change(old, new)
 
    def _recurs_replace_child_with(self, dict_, child_a, child_b):
        if isinstance(dict_, dict):
            for key in dict_:
                value = dict_[key]
                if value == child_a:
                     dict_[key] = child_b
                elif isinstance(value, dict):
                    self._recurs_replace_child_with(value, child_a, child_b)
            
    def _on_change(self, old, new):
        """Called everytime the dict is mutated (i.e. my_dict['foo'] = 'bar')
        Overwrite this function with what you want your callback to be
        
        Args
        ----
        old: a copy of the dictionary object before the mutation
        new: a copy of the dictionary object after the mutation
        """
        pass

class ObservableDict(Dict):
    """Mimics the traitlets `Dict` class, but will fire a `change` event when
    a dictionary object is updated (new value added, etc.). Can get hooked up
    with traitlets/ipywidget `observe()` func and some other thigns"""

    _parent = None
    def get(self, obj, cls):
        self._parent = obj
        orig_out = super().get(obj, cls)
        new_out = observabledict(orig_out)
        new_out._on_change = self._on_change
        self._parent.set_trait(self.name, new_out)
        return new_out

    def _on_change(self, old, new):
        self._parent.notify_change({'name': self.name,
                                    'old': old,
                                    'new': new,
                                    'owner': self._parent,
                                    'type': 'change'})



from .._utils.env import LAMBDA_TEXT_CLASSIFICATION

if not LAMBDA_TEXT_CLASSIFICATION:
    from ._ssd import SingleShotDetector
    from ._inferencing import _DynamicSSD
    from ._unet import UnetClassifier
    from ._retinanet import RetinaNet
    from ._classifier import FeatureClassifier
    from ._pspnet import PSPNetClassifier
    from ._maskrcnn import MaskRCNN
    from ._deeplab import DeepLab
    from ._pointcnnseg import PointCNN
    from ._yolov3 import YOLOv3
    from ._layer_learner import FullyConnectedNetwork
    from ._machine_learning import MLModel
    from ._model_extension import ModelExtension
    from ._faster_rcnn import FasterRCNN
    from ._superres import SuperResolution
    from ._hed import HEDEdgeDetector
    from ._bdcn import BDCNEdgeDetector
    from ._image_captioner import ImageCaptioner
    from ._cyclegan import CycleGAN
    from ._tsmodel import TimeSeriesModel
    from ._multi_task_road_extractor import MultiTaskRoadExtractor
    from ._change_detector import ChangeDetector
    from ._pix2pix import Pix2Pix




from typing import Collection
import fastai


precondition = True

if fastai.__version__=='1.0.60':

    import fastai.vision.image
    from fastai.vision.image import Image, plt, image2np
    from .common import ArcGISMSImage
    def show_image(img:Image, ax:plt.Axes=None, figsize:tuple=(3,3), hide_axis:bool=True, cmap:str='binary',
                    alpha:float=None, **kwargs)->plt.Axes:
        "Display `Image` in notebook."
        if ax is None: fig,ax = plt.subplots(figsize=figsize)
        xtr = dict(cmap=cmap, alpha=alpha, **kwargs)
        if isinstance(img, ArcGISMSImage):
            img.show(ax=ax)
        else:
            ax.imshow(image2np(img.data), **xtr) if (hasattr(img, 'data')) else ax.imshow(img, **xtr)
        if hide_axis: ax.axis('off')
        return ax
    #
    fastai.vision.learner.show_image = show_image
    fastai.vision.image.show_image = show_image


    import fastai.data_block
    from fastai.core import array
    def process(self, ds:Collection):
        ds.items = array([self.process_one(item) for item in ds.items], dtype=object)
    fastai.data_block.PreProcessor.process = process


